"""Eargait: The Gait and Movement Analysis Package unsing Ear-Worn IMU sensors."""
from eargait.event_detection.base_event_detection import BaseEventDetection
from eargait.event_detection.diao_adapted_event_detection import DiaoAdaptedEventDetection
from eargait.event_detection.diao_event_detection import DiaoEventDetection

__all__ = ["BaseEventDetection", "DiaoEventDetection", "DiaoAdaptedEventDetection"]



"""Set of functions to estimate temporal gait parameters."""
import warnings
from typing import Dict, Union

import numpy as np
import pandas as pd

from eargait.utils.helper_datatype import EventList, SingleSensorEventList, is_event_list


def get_temporal_params(event_list: EventList, sampling_rate_hz: int) -> Union[Dict, pd.DataFrame]:
    """Get temporal parameters of event list.

    Parameters
    ----------
    event_list
        Detected event list
    sampling_rate_hz
        Sampling rate of data
    event_consistency:
        Delete temporal parameters of events which are not adjecent events in gait sequence.

    Returns
    -------
    temporal_params
        pd.DataFrames containing temporal gait parameters

    """
    kind = is_event_list(event_list)
    if kind == "single":
        return _get_single_temporal_params(event_list, sampling_rate_hz)
    temporal_params = {}
    for pos in event_list.keys():
        temporal_params[pos] = _get_single_temporal_params(event_list[pos], sampling_rate_hz)
    return temporal_params


def get_average_temporal_params(temporal_params: Union[Dict, pd.DataFrame]) -> Union[Dict, pd.DataFrame]:
    """Average temporal gait events (event, stance and swing time)."""
    return get_average_params(temporal_params)


def _get_single_temporal_params(
    event_list: SingleSensorEventList,
    sampling_rate_hz: int,
) -> pd.DataFrame:

    df = pd.DataFrame(index=event_list.index, columns=["stride_time", "stance_time", "swing_time"])
    for lateral in ["ipsilateral", "contralateral"]:
        tmp = event_list.loc[event_list.side == lateral]
        df.loc[event_list.side == lateral] = _get_single_temporal_params_single_lateral(tmp, sampling_rate_hz)
    for col in df.columns:
        df[col] = df[col].astype(float)
    df["side"] = event_list.side
    return df


def _get_single_temporal_params_single_lateral(
    event_list: SingleSensorEventList, sampling_rate_hz: int
) -> pd.DataFrame:
    df = pd.DataFrame(index=event_list.index, columns=["stride_time", "stance_time", "swing_time"])
    df["stride_time"] = pd.Series(
        (event_list.ic.diff().shift(-1).iloc[0:-1] / sampling_rate_hz).to_numpy(), index=df.index[0:-1]
    )
    stance = pd.Series(
        (event_list.tc.iloc[1:].to_numpy() - event_list.ic.iloc[0:-1].to_numpy()) / sampling_rate_hz,
        index=df.index[0:-1],
    )
    df["swing_time"] = pd.Series(
        np.array((event_list.ic.iloc[1::] - event_list.tc.iloc[1::])) / sampling_rate_hz, index=df.index[0:-1]
    )
    df["stance_time"] = stance
    _sanity_check_temporal_parameters(df)
    return df


def get_average_params(params: Union[Dict, pd.DataFrame]) -> Union[Dict, pd.DataFrame]:
    """Average temporal gait events (stride, stance and swing time)."""
    if not isinstance(params, Dict):
        if "side" in params.columns:
            params = params.drop(columns="side").astype(float)
        return params.describe().loc[["mean", "std"]]
    # else: multi sensor
    avaerage_params = {}
    for pos, param in params.items():  # noqa
        if "side" in param.columns:
            param = param.drop(columns="side").astype(float)
        avaerage_params[pos] = param.describe().loc[["mean", "std"]]
    return avaerage_params


def get_average_spatial_params(spatial_params: Union[Dict, pd.DataFrame]) -> Union[Dict, pd.DataFrame]:
    """Average spatial parametets (stride_length, step_length)."""
    return get_average_params(spatial_params)


def combine_spatial_temporal(temporal, spatial):
    if not isinstance(temporal, Dict):
        return _combine_single_spatial_temporal(temporal, spatial)
    temporal_spatial = {}
    for sensor in temporal.keys():
        temporal_spatial[sensor] = _combine_single_spatial_temporal(temporal[sensor], spatial[sensor])
    return temporal_spatial


def get_average_spatiotemp_params(spatiotemp_params: Union[Dict, pd.DataFrame]) -> Union[Dict, pd.DataFrame]:
    """Average spatiotemporal parameter."""
    return get_average_params(spatiotemp_params)


def _combine_single_spatial_temporal(temp, spat):
    assert temp.side.equals(spat.side)
    spat = spat.drop(columns=["side"])
    return pd.merge(temp, spat, left_index=True, right_index=True)


def _check_stride_consistency(df: pd.DataFrame, event_list_type):
    """Check for sequential stride index and removes entries with missing index inbetween."""
    if event_list_type == "upper_body_sensor_unilateral":
        diff_index = (df.index.to_series().diff() == 1) | (df.index.to_series().diff() > 2)
    else:
        diff_index = df.index.to_series().diff() > 1
    df[diff_index] = np.nan
    return df


def _sanity_check_temporal_parameters(tmp: pd.DataFrame):
    df = tmp.copy()
    df = df.dropna(axis=0)
    assert np.round((df["swing_time"] + df["stance_time"]), 4).equals(np.round(df["stride_time"], 4))
    info = df.describe()
    if info.loc["mean"].stride_time < 0.8 or info.loc["mean"].stride_time > 1.5:
        message = "Unrealistic average stride time of  {:.2f}s".format(info.loc["mean"].stride_time)  # noqa
        warnings.warn(message)
    if info.loc["mean"].swing_time < 0.25 or info.loc["mean"].swing_time > 0.8:
        message = "Unrealistic average swing time of  {:.2f}s".format(info.loc["mean"].swing_time)  # noqa
        warnings.warn(message)
    if info.loc["mean"].stance_time < 0.4 or info.loc["mean"].stance_time > 1.0:
        message = "Unrealistic average stance time of  {:.2f}s".format(info.loc["mean"].stance_time)  # noqa
        warnings.warn(message)



"""A set of functions to rotate and transform ear-worn sensor data."""
from typing import Dict, Union

import numpy as np
from nilspodlib import SyncedSession
from signialib import Session

from eargait.utils.consts import SF_ACC, SF_COLS
from eargait.utils.helper_gaitmap import (
    MultiSensorData,
    SensorData,
    align_dataset_to_gravity,
    convert_to_fbf,
    is_multi_sensor_data,
    rotate_dataset,
    rotation_from_angle,
)


def convert_ear_to_ebf(session: Union[Session, SyncedSession]) -> MultiSensorData:
    """Convert sensor data from ear worn sensors into the ear body frame (ebf).

    Parameters
    ----------
    session :
            Recorsing session either of type SigniaSession or Nilspod SyncSession.

    Return
    ------
    MutliSensorData
            Contains gyroscope and acceleration data in the ear body frame

    Notes
    -----
    This funktion can only be used if session data is in hearing aid frame. A different rotation is chosen if
    firmware version is D12. See user guide, coordinate systems for more information.

    """
    dataset_sf = convert_ear_to_esf(session)
    return convert_esf_to_ebf(dataset_sf)


def aling_gravity_and_convert_ear_to_ebf(session: Union[Session, SyncedSession]) -> SensorData:
    """Convert sensor data from hearing aid frame into the ear body frame (ebf).

    Parameters
    ----------
    session :
            Recording session either of type SigniaSession or Nilspod SyncSession.

    Return
    ------
    MutliSensorData
            Contains gyroscope and acceleration data in the ear body frame

    Notes
    -----
    This funktion can only be used if session data is in hearing aid frame. A different rotation is chosen if
    firmware version is D12. See user guide, coordinate systems for more information.

    """
    dataset_sf = convert_ear_to_esf(session)
    aligned = {}
    if len(dataset_sf) == 1:
        sensor_pos = list(dataset_sf.keys())[0]
        aligned[sensor_pos] = align_dataset_to_gravity(dataset_sf[sensor_pos], session.info.sampling_rate_hz[0])
    else:
        aligned = align_dataset_to_gravity(dataset_sf, session.info.sampling_rate_hz[0])
    return convert_esf_to_ebf(aligned)


def convert_ear_to_esf(session: Union[Session, SyncedSession]) -> SensorData:
    """Convert sensor data from hearing aid frame (haf) into the ear sensor frame (esf).

    Parameters
    ----------
    session :
            Recording session of type SigniaSession

    Return
    ------
    SensorData
            Contains gyroscope and acceleration data in the ear body frame

    Notes
    -----
    This funktion can only be used if session data is in hearing aid frame. A different rotation is chosen if
    firmware version is D12. See user guide, coordinate systems for more information.

    """
    # data into sensor_data (single or multi)
    dataset, gyr_avail = get_ear_multi_sensor_data(session)

    # rotate to sensor frame
    rotation = _get_rotation(session)
    dataset_sf = rotate_dataset(dataset, rotation, check_gyr=gyr_avail)
    is_multi_sensor_data(dataset_sf, frame="sensor", raise_exception=True, check_gyr=gyr_avail)
    # rotate to body frame and return
    return dataset_sf


def convert_esf_to_ebf(datasets: SensorData) -> SensorData:
    """Convert sensor data from ear sensor frame (esf) into the ear body frame (ebf).

    Parameters
    ----------
    datasets:
            Dataset

    Return
    ------
    SensorData
            Contains gyroscope and acceleration data in the ear body frame

    Notes
    -----
    See user guide, coordinate systems for more information.

    """
    if len(datasets) == 1:
        sensor_pos = list(datasets.keys())[0]
        if sensor_pos == "left_sensor":
            return convert_to_fbf(datasets, left=["left_sensor"])
        return convert_to_fbf(datasets, right=["right_sensor"])
    return convert_to_fbf(datasets, left=["left_sensor"], right=["right_sensor"])


def _get_rotation(session: Union[SyncedSession, Session, str]) -> Dict:
    if isinstance(session, Session) or session == "signia":

        if "D12" in session.info.version_firmware[0]:
            if "BMA400" == session.info.imu_sensor_type[0]:
                rot_matrices = _get_rot_matrix_d12_bma400()
                print("bma 400 and D12")
            else:
                rot_matrices = _get_rot_matrix_d12()
                print("D12")
        else:
            rot_matrices = _get_rot_matrix_default()
            print("D11, default")

        rot = {}
        for side in session.info.sensor_position:
            side = side.split("_")[1]
            rot[side + "_sensor"] = rot_matrices[side]
    elif isinstance(session, SyncedSession) or session == "nilspod":
        left_rot = rotation_from_angle(np.array([1, 0, 0]), np.deg2rad(-90))
        right_rot = rotation_from_angle(np.array([0, 0, 1]), np.deg2rad(180)) * rotation_from_angle(
            np.array([1, 0, 0]), np.deg2rad(-90)
        )
        rot = {"left_sensor": left_rot, "right_sensor": right_rot}
    return rot


def _get_rot_matrix_default():
    rot_matrices = {}
    rot_matrices["left"] = rotation_from_angle(np.array([0, 0, 1]), np.deg2rad(-180)) * rotation_from_angle(
        np.array([0, 1, 0]), np.deg2rad(-90)
    )
    rot_matrices["right"] = rotation_from_angle(np.array([0, 0, 1]), np.deg2rad(-180)) * rotation_from_angle(
        np.array([0, 1, 0]), np.deg2rad(-90)
    )
    return rot_matrices


def _get_rot_matrix_d12():
    rot_matrices = {}
    rot_matrices["left"] = rotation_from_angle(np.array([1, 0, 0]), np.deg2rad(-90)) * rotation_from_angle(
        np.array([0, 1, 0]), np.deg2rad(-90)
    )
    rot_matrices["right"] = rotation_from_angle(np.array([1, 0, 0]), np.deg2rad(-90)) * rotation_from_angle(
        np.array([0, 1, 0]), np.deg2rad(-90)
    )
    return rot_matrices


def _get_rot_matrix_d12_bma400():
    rot_matrices = {}
    rot_matrices["left"] = rotation_from_angle(np.array([0, 1, 0]), np.deg2rad(90))
    rot_matrices["right"] = rotation_from_angle(np.array([0, 1, 0]), np.deg2rad(90))
    return rot_matrices


def get_ear_multi_sensor_data(session: Union[SyncedSession, Session]) -> Dict:
    if isinstance(session, SyncedSession):
        left = session.get_dataset_by_id("157e").imu_data_as_df()
        right = session.get_dataset_by_id("2541").imu_data_as_df()
        left, right = left[SF_COLS], right[SF_COLS]
        return {"left_sensor": left, "right_sensor": right}
    res = {}
    for dataset in session.datasets:
        pos = dataset.info.sensor_position.split("_")[1]
        if session.gyro[0]:
            res[pos + "_sensor"] = dataset.data_as_df()[SF_COLS]
            gyr_avail = True
        else:
            res[pos + "_sensor"] = dataset.data_as_df()[SF_ACC]
            gyr_avail = False
    return res, gyr_avail



"""Helper functions to load example data."""

from pathlib import Path
from typing import Tuple

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from PIL import Image

from eargait.utils.helper_gaitmap import MultiSensorData

HERE = Path(__file__).parent.parent.parent


def get_mat_example_data_path() -> Path:
    # data directory
    data_dir = HERE.joinpath("example_data/mat_files/normal")
    return data_dir


def get_txt_example_data_path() -> Path:
    # data directory
    data_dir = HERE.joinpath("example_data/txt_file/08-03-2023_11-12-51_txt_example_no_gait_data.txt")
    return data_dir


def get_example_data() -> Tuple[MultiSensorData, int]:
    # returns example data for a single walking,  50Hz
    # sensor data is aligned, calibrated
    csv_data = pd.read_csv(
        HERE.joinpath("example_data", "aligned_calibrated_earframe_50Hz_walking_data_single_walkingbout.csv")
    )
    left = csv_data.loc[csv_data.side == "left"].drop(columns=["side"]).set_index("n_samples")
    right = csv_data.loc[csv_data.side == "right"].drop(columns=["side"]).set_index("n_samples")
    return {"left_sensor": left, "right_sensor": right}, 50


def plot_image(image_path):
    pil_im = Image.open(image_path)
    im_array = np.asarray(pil_im)
    plt.axis("off")
    plt.imshow(im_array)
    plt.show()



"""A set of functions to load a recording of IMU sensors integrated into signia hearing aids."""

import warnings
from pathlib import Path

import numpy as np
from signialib import SIGNIA_CAL_PATH, Session


def load(path: str, target_sample_rate_hz: int = 0, skip_calibration=False):
    """Load motion data recorded using signia hearing aids.

    Parameters
    ----------
    path
        Local path of folder containing the recording.

    target_sample_rate_hz
        The target sampling rate of the data. The target sampling rate only has to be provided, if data
        should be resampled to a lower sampling rate.
        Default 0 means, no resampling.

    skip_calibration
        If True, (Ferraris) calibration is skipped. It's strongly recommended to use calibration.
        Calibration file for each sensor is stored in signia lib.

    Returns
    -------
    session
        Signia Session

    Notes
    -----
    This function can only be applied to dataset recorded using the Signia Hearing Aids.

    """
    path = Path(path)
    if path.is_file():
        session = Session.from_file_path(path)
    else:
        session = Session.from_folder_path(path)
    if target_sample_rate_hz != 0:
        _check_valid_resample_rate(session.info.sampling_rate_hz[0], target_sample_rate_hz)
    else:
        target_sample_rate_hz = session.info.sampling_rate_hz[0]
    if skip_calibration:
        warnings.warn("Calibration was skipped. Calibration is strongly recommended.")
        if len(session.datasets) == 2:
            session = session.align_to_syncregion()
        else:
            warnings.warn("Single dataset in session, alignment not necessary.")
        session = session.resample(target_sample_rate_hz)
    else:
        session = session.align_calib_resample(resample_rate_hz=target_sample_rate_hz, calib_path=SIGNIA_CAL_PATH)
    return session


def _check_valid_resample_rate(orig_sample_rate_hz, target_rate_hz):
    if np.mod(orig_sample_rate_hz, target_rate_hz) != 0:
        raise ValueError(
            "Please choose valid target sample rate. Has to be divivend of original sample rate {}Hz".format(  # noqa
                int(orig_sample_rate_hz)
            )
        )
    return np.mod(orig_sample_rate_hz, target_rate_hz)



"""The event detection algorithm by Diao et al, 2020 [1] with adaptations."""
import warnings
from typing import Tuple, TypeVar

import numpy as np
import pandas as pd
from pyts.decomposition import SingularSpectrumAnalysis

from eargait.event_detection.diao_event_detection import DiaoEventDetection

Self = TypeVar("Self", bound="DiaoAdaptedEventDetection")


class DiaoAdaptedEventDetection(DiaoEventDetection):
    """Find gait events in the IMU raw signal based on adapted Diao algorithm.

    DiaoEventDetection uses a Singuar Spectrum Analysis to decompose acceleration data.
    The peaks in the dominant ozcillation of the SI axis are consindered to be the initial contact (IC).
    Ipsi and contralateral sides are assigned by looking at the adjecent sample in the ML axis.
    Terminate contacts (TC) are determined based the next minima or maxima in the ML axis.
    More information on the algorithm can be found in the paper by Diao et al. [1].

    To Do: report adaptations

    Parameters
    ----------
    data
        The IMU data passed to the `detect` method.
    sampling_rate_hz
        The sampling rate of the data
    ssa
        Singular Spectrum Analysis (SSA).
    window_length
        The window length for the SSA.



    Attributes
    ----------
    event_list_
        Event list or dictionary with such values.
        The result of the `detect` method holding all temporal gait events. A stride is defined
        such that terminate contact is follwed by an initial contact of the same foot.
        Gait sequences with no steady alternating sequence of ipsi- and contralateral strides are removed.
    nonconsistent_event_list_
        Event list with non consistent gait events. Non-consistent gait event are stride that do not follow the
        stride defintion as described in 'event_list_', e.g. initial contact is before the terminate contact.

    window_length

    Examples
    --------
    Get gait events from single sensor signal

    >>> event_detection = DiaoAdaptedEventDetection()
    >>> event_detection.detect(data=data, sample_rate_hz=200.0)
            ic      tc       side
    s_id
    0      651.0    584.0    ipsilateral
    1      839.0    802.0    contralateral
    2      1089.0   1023.0   ipsilateral
    ...


    Notes
    -----
    terminal contact (`tc`), often also referred to as toe off:
        At `tc` the movement of the ankle joint changes from a plantar flexion to a dorsal extension in the sagittal
        plane..

    initial contact (`ic`), often also referred to as heel strike:
        At `ic` the foot decelerates rapidly when the foot hits the ground.

    ipsilateral and contralateral:
        In contrast to foot worn sensors, ear worn sensors contain gait events from left and right foot.
        An event of the foot situated at same body side as the ear worn sensor is referred to as ipsilateral event.
        An event of the foot situated at the oppisite body side if referred to as contralateral event.
        A gait sequence is a steady alternating sequence of ipsi- and contralateral events.


    [1] Diao, Y., Ma, Y., Xu, D., Chen, W., & Wang, Y. (2020). A novel gait parameter estimation method for healthy
    adults and postoperative patients with an ear-worn sensor. Physiological measurement, 41(5), 05NT01.

    """

    def __init__(
        self,
        sample_rate_hz: int,
        window_length: int = None,
    ):
        super().__init__(sample_rate_hz=sample_rate_hz, window_length=window_length)

    def _initialize_ssa(self) -> Self:
        self.ssa = SingularSpectrumAnalysis(
            window_size=self.window_length, groups=[[0], [1], [2], np.arange(3, self.window_length, 1)]
        )

    def _find_all_events(
        self, acc: pd.DataFrame, ssa: SingularSpectrumAnalysis, sample_rate_hz: float
    ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """Find events in provided data by looping over single strides."""
        acc_si = acc["acc_si"].to_numpy().reshape(1, -1)
        acc_ssa_si = ssa.fit_transform(acc_si)
        acc_ml = acc["acc_ml"].to_numpy().reshape(1, -1) * -1  # changed
        acc_ssa_ml = ssa.fit_transform(acc_ml)

        ic, ic_sides = self._detect_ic(acc_ssa_si[1], acc_ssa_ml[1], sample_rate_hz)
        tc, tc_sides = self._detect_tc(acc_ssa_ml[2] + acc_ssa_ml[1] + acc_ssa_ml[3], ic, ic_sides)

        ic += acc.index[0]
        tc += acc.index[0]

        if np.abs(acc.shape[0] / (sample_rate_hz * ic.shape[0]) - 0.5) > 0.15:
            # print(np.abs(acc.shape[0]/(sample_rate_hz*ic.shape[0])-0.5))
            msg = (
                f"Walking bout length and number of ICs are unrealistic: "
                f"{acc.shape[0] / sample_rate_hz:.2f}s and {ic.shape[0]} steps"
            )
            warnings.warn(msg, UserWarning)

        return (
            pd.DataFrame.from_dict({"ic": ic, "side": ic_sides}),
            pd.DataFrame.from_dict({"tc": tc, "side": tc_sides}),
        )



"""Constants."""

#: The default names of the Gyroscope columns in the sensor frame
SF_GYR = ["gyr_x", "gyr_y", "gyr_z"]
#: The default names of the Accelerometer columns in the sensor frame
SF_ACC = ["acc_x", "acc_y", "acc_z"]
#: The default names of all columns in the sensor frame
SF_COLS = [*SF_ACC, *SF_GYR]

#: The default names of the Gyroscope columns in the body frame
BF_GYR = ["gyr_pa", "gyr_ml", "gyr_si"]
#: The default names of the Accelerometer columns in the body frame
BF_ACC = ["acc_pa", "acc_ml", "acc_si"]
#: The default names of all columns in the body frame
BF_COLS = [*BF_ACC, *BF_GYR]



"""Set of helper functions to to deal with different data types."""
from typing import Dict, Hashable, Union

import pandas as pd
from typing_extensions import Literal

from eargait.utils.helper_gaitmap import ValidationError, is_stride_list

SingleSensorEventList = pd.DataFrame
MultiSensorEventList = Union[pd.DataFrame, Dict[Union[Hashable, str], SingleSensorEventList]]
EventList = Union[SingleSensorEventList, MultiSensorEventList]


def is_event_list(event_list: EventList) -> Literal["single", "multi"]:
    """Check if an object is a valid multi-sensor or single-sensor event list.

    This function utilizes the is_stride_list function by gaitmap.
    For more information, please refer to: eargait.utils.helper_gaitmap.is_stride_list

    Parameters
    ----------
    event_list
        The object that should be tested

    Returns
    -------
    dataset_type
        "single" in case of a single-sensor event list, "multi" in case of a multi-sensor event list.
        In case it is neither, a ValidationError is raised.

    """
    return is_stride_list(event_list)


def get_event_list_type(
    event_list: EventList,
) -> Literal["lower_body_sensor", "upper_body_sensor_bilateral", "upper_body_sensor_unilateral"]:
    """Check if an object is a valid multi-sensor or single-sensor event list.

    In case one of the two checks is successful, a string is returned, which type of dataset the input is.
    Otherwise a descriptive error is raised

    Parameters
    ----------
    event_list
        The object that should be tested

    Returns
    -------
    dataset_type
        "lower_body_sensor" in case of a event list by e.g. foot-mounted sensors,
        "upper_body_sensor_bilateral" in case of a event list by a sensor mounted to the upper body
        (e.g. ear, head, lower back) including events of both legs,
        "upper_body_sensor_unilateral" in case of a event list by a sensor mounted to the upper body
        (e.g. ear, head, lower back) including events of single leg.
        In case it is neither, a ValidationError is raised.

    """
    kind = is_event_list(event_list)
    check_bilateral(event_list)

    if kind == "single":
        event_list_type = _get_event_list_type_single_sensor_list(event_list)
    else:
        event_list_type_list = []
        for sensor_pos in event_list.keys():
            event_list_type_list.append(_get_event_list_type_single_sensor_list(event_list[sensor_pos]))
        if len(set(event_list_type_list)) > 1:
            raise ValidationError(
                "The passed object appears to be neither a multi-sensor event list. "
                "However, the lists appear to be of two different kinds:\n\n"
                f"{str(event_list_type_list)}\n\n"
            )
        event_list_type = event_list_type_list[0]
    return event_list_type


def _get_event_list_type_single_sensor_list(event_list: SingleSensorEventList):
    if "side" in event_list.columns:
        if event_list.side.nunique() == 1:
            return "upper_body_sensor_unilateral"
        if event_list.side.nunique() == 2:
            return "upper_body_sensor_bilateral"
        return None
    return "lower_body_sensor"


def check_bilateral(event_list: EventList):
    kind = is_event_list(event_list)

    if kind == "single":
        _check_bilateral_single(event_list)
    else:
        for events in event_list.values():
            _check_bilateral_single(events)


def _check_bilateral_single(event_list: EventList):
    if "side" in event_list.columns:
        side_set = set(event_list.side.unique())
        if not side_set.issubset(set(["ipsilateral", "contralateral"])):
            raise ValueError(
                "Side of bilateral sensor are not in [ipsilateral, contralaeral]", event_list.side.unique()
            )



"""This is a collection of helpers that were copied from MaDLab internal python package gaitmap.
The original copyright belongs to MaD-DiGait group.
All original authors agreed to have the respective code snippets published in this way.
In case any problems are detected, these changes should be upstreamed to the respective internal libraries.
"""

import warnings
from typing import Any, Dict, Hashable, Iterable, List, Optional, Tuple, Type, TypeVar, Union

import numpy as np
import pandas as pd
import tpcp
from scipy.spatial.transform import Rotation
from typing_extensions import Literal

from eargait.utils.consts import *

GRAV_VEC = np.array([0.0, 0.0, 9.81])
METRIC_FUNCTION_NAMES = Literal["maximum", "variance", "mean", "median"]

BaseType = TypeVar("BaseType", bound="_BaseSerializable")

SingleSensorData = pd.DataFrame
SingleSensorStrideList = pd.DataFrame

MultiSensorData = Union[pd.DataFrame, Dict[Union[Hashable, str], SingleSensorData]]
SensorData = Union[SingleSensorData, MultiSensorData]

MultiSensorStrideList = Dict[Union[Hashable, str], pd.DataFrame]
StrideList = Union[SingleSensorStrideList, MultiSensorStrideList]

_ALLOWED_FRAMES = ["any", "body", "sensor"]
_ALLOWED_FRAMES_TYPE = Literal["any", "body", "sensor"]

#: Sensor to body frame conversion for the left foot
FSF_FBF_CONVERSION_LEFT = {
    "acc_x": (1, "acc_pa"),
    "acc_y": (1, "acc_ml"),
    "acc_z": (-1, "acc_si"),
    "gyr_x": (-1, "gyr_pa"),
    "gyr_y": (-1, "gyr_ml"),
    "gyr_z": (-1, "gyr_si"),
}

#: Sensor to body frame conversion for the right foot
FSF_FBF_CONVERSION_RIGHT = {
    "acc_x": (1, "acc_pa"),
    "acc_y": (-1, "acc_ml"),
    "acc_z": (-1, "acc_si"),
    "gyr_x": (1, "gyr_pa"),
    "gyr_y": (-1, "gyr_ml"),
    "gyr_z": (1, "gyr_si"),
}


class ValidationError(Exception):
    """An error indicating that data-object does not comply with the guidelines."""


class _BaseSerializable(tpcp.BaseTpcpObject):
    @classmethod
    def _get_subclasses(cls: Type[BaseType]):
        for subclass in cls.__subclasses__():
            yield from subclass._get_subclasses()
            yield subclass

    @classmethod
    def _find_subclass(cls: Type[BaseType], name: str) -> Type[BaseType]:
        for subclass in _BaseSerializable._get_subclasses():
            if subclass.__name__ == name:
                return subclass
        raise ValueError("No algorithm class with name {} exists".format(name))


class BaseAlgorithm(tpcp.Algorithm, _BaseSerializable):
    """Base class for all algorithms.

    All type-specific algorithm classes should inherit from this class and need to

    1. overwrite `_action_method` with the name of the actual action method of this class type
    2. implement a stub for the action method

    Attributes
    ----------
    _action_method
        The name of the action method used by the Childclass

    """


class BaseEventDetection(BaseAlgorithm):
    """Base class for all event detection algorithms."""

    _action_methods = ("detect",)

    def detect(self: BaseType, data: SensorData, stride_list: StrideList, sampling_rate_hz: float) -> BaseType:
        """Find gait events in data within strides provided by roi_list."""
        raise NotImplementedError("Needs to be implemented by child class.")


def invert_result_dictionary(
    nested_dict: Dict[Hashable, Dict[TypeVar("_HashableVar", Hashable, str), Any]]
) -> Dict[TypeVar("_HashableVar", Hashable, str), Dict[Hashable, Any]]:
    """Invert result dictionaries that are obtained from multi sensor results.

    This method expects a two level dictionary and flips the levels.
    This means that if a value can be accessed as `nested_dict[k1][k2] = v` in the input, it can be accessed as
    `output_dict[k2][k1] = v`.

    Examples
    --------
    >>> in_dict = {"level_1_1": {"level_2_1": "val_2_1", "level_2_2": "val_2_2"}, "level_1_2": {"level_2_3": "val_2_3"}}
    >>> from pprint import pprint
    >>> pprint(invert_result_dictionary(in_dict))
    {'level_2_1': {'level_1_1': 'val_2_1'},
     'level_2_2': {'level_1_1': 'val_2_2'},
     'level_2_3': {'level_1_2': 'val_2_3'}}

    """
    out: Dict[TypeVar("_HashableVar", Hashable, str), Dict[Hashable, Any]] = dict()
    for ok, ov in nested_dict.items():
        for k, v in ov.items():
            nested = out.setdefault(k, dict())
            nested[ok] = v
    return out


def set_params_from_dict(obj: Any, param_dict: Dict[str, Any], result_formatting: bool = False):
    """Update object attributes from dictionary.

    The object will be updated inplace.

    Parameters
    ----------
    obj
        The gaitmap obj to update
    param_dict
        The dictionary of new values to set/update
    result_formatting
        If True all keys will get a trailing "_", if they don't have one already.
        This marks them as "results" based on the gaitmap guidelines.

    """
    for k, v in param_dict.items():
        if result_formatting is True:
            if not k.endswith("_"):
                k += "_"
        setattr(obj, k, v)


def get_multi_sensor_names(dataset: MultiSensorData):
    """Get the list of sensor names from a multi-sensor dataset.

    .. warning:
        This will not check, if the input is actually a multi-sensor dataset

    Notes
    -----
    The keys are not guaranteed to be ordered.

    """
    return _get_multi_sensor_data_names(dataset=dataset)


def _get_multi_sensor_data_names(dataset: Union[dict, pd.DataFrame]):
    if isinstance(dataset, pd.DataFrame):
        keys = dataset.columns.unique(level=0)
    else:
        # In case it is a dict
        keys = dataset.keys()

    return keys


def _assert_is_dtype(obj, dtype: Union[type, Tuple[type, ...]]):
    """Check if an object has a specific dtype."""
    if not isinstance(obj, dtype):
        raise ValidationError("The dataobject is expected to be one of ({},). But it is a {}".format(dtype, type(obj)))


def _assert_has_multindex_cols(df: pd.DataFrame, nlevels: int = 2, expected: bool = True):
    """Check if a pd.DataFrame has a multiindex as columns.

    Parameters
    ----------
    df
        The dataframe to check
    nlevels
        If MultiIndex is expected, how many level should the MultiIndex have
    expected
        If the df is expected to have a MultiIndex or not

    """
    has_multiindex = isinstance(df.columns, pd.MultiIndex)
    if has_multiindex is not expected:
        if expected is False:
            raise ValidationError(
                "The dataframe is expected to have a single level of columns. "
                "But it has a MultiIndex with {} levels.".format(df.columns.nlevels)
            )
        raise ValidationError(
            "The dataframe is expected to have a MultiIndex with {} levels as columns. "
            "It has just a single normal column level.".format(nlevels)
        )
    if has_multiindex is True:
        if not df.columns.nlevels == nlevels:
            raise ValidationError(
                "The dataframe is expected to have a MultiIndex with {} levels as columns. "
                "It has a MultiIndex with {} levels.".format(nlevels, df.columns.nlevels)
            )


def _assert_has_columns(df: pd.DataFrame, columns_sets):
    """Check if the dataframe has at least all columns sets.

    Examples
    --------
    >>> df = pd.DataFrame()
    >>> df.columns = ["col1", "col2"]
    >>> _assert_has_columns(df, [["other_col1", "other_col2"], ["col1", "col2"]])
    >>> # This raises no error, as df contains all columns of the second set

    """
    columns = df.columns
    result = False
    for col_set in columns_sets:
        result = result or all(v in columns for v in col_set)

    if result is False:
        if len(columns_sets) == 1:
            helper_str = "columns: {}".format(columns_sets[0])
        else:
            helper_str = "one of the following sets of columns: {}".format(columns_sets)
        raise ValidationError(
            "The dataframe is expected to have {}. Instead it has the following columns: {}".format(
                helper_str, list(df.columns)
            )
        )


def _get_expected_dataset_cols(
    frame: Literal["sensor", "body"], check_acc: bool = True, check_gyr: bool = True
) -> List:
    expected_cols = []
    if frame == "sensor":
        acc = SF_ACC
        gyr = SF_GYR
    elif frame == "body":
        acc = BF_ACC
        gyr = BF_GYR
    else:
        raise ValueError('`frame must be one of ["sensor", "body"]')
    if check_acc is True:
        expected_cols.extend(acc)
    if check_gyr is True:
        expected_cols.extend(gyr)
    return expected_cols


def is_single_sensor_data(
    data: SingleSensorData,
    check_acc: bool = True,
    check_gyr: bool = True,
    frame: _ALLOWED_FRAMES_TYPE = "any",
    raise_exception: bool = False,
) -> Optional[bool]:
    """Check if an object is a valid dataset following all conventions.

    A valid single sensor dataset is:

    - a :class:`pandas.DataFrame`
    - has only a single level of column indices that correspond to the sensor (or feature) axis that are available.

    A valid single sensor dataset in the body frame additionally:

    - contains all columns listed in :obj:`SF_COLS <gaitmap.utils.consts.SF_COLS>`

    A valid single sensor dataset in the sensor frame additionally:

    - contains all columns listed in :obj:`BF_COLS <gaitmap.utils.consts.BF_COLS>`

    Parameters
    ----------
    data
        Object that should be checked
    check_acc
        If the existence of the correct acc columns should be checked
    check_gyr
        If the existence of the correct gyr columns should be checked
    frame
        The frame the dataset is expected to be in.
        This changes which columns are checked for.
        In case of "any" a dataset is considered valid if it contains the correct columns for one of the two frames.
        If you just want to check the datatype and shape, but not for specific column values, set both `check_acc` and
        `check_gyro` to `False`.
    raise_exception
        If True an exception is raised if the object does not pass the validation.
        If False, the function will return simply True or False.

    See Also
    --------
    gaitmap.utils.dataset_helper.is_multi_sensor_data: Explanation and checks for multi sensor datasets

    """
    if frame not in _ALLOWED_FRAMES:
        raise ValueError("The argument `frame` must be one of {}".format(_ALLOWED_FRAMES))
    try:
        _assert_is_dtype(data, pd.DataFrame)
        _assert_has_multindex_cols(data, expected=False)

        if frame == "any":
            _assert_has_columns(
                data,
                [
                    _get_expected_dataset_cols("sensor", check_acc=check_acc, check_gyr=check_gyr),
                    _get_expected_dataset_cols("body", check_acc=check_acc, check_gyr=check_gyr),
                ],
            )
        else:
            _assert_has_columns(data, [_get_expected_dataset_cols(frame, check_acc=check_acc, check_gyr=check_gyr)])

    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be SingleSensorData. "
                "The validation failed with the following error:\n\n{}".format(str(e))
            ) from e
        return False
    return True


def _assert_multisensor_is_not_empty(obj: Union[pd.DataFrame, Dict]):
    sensors = _get_multi_sensor_data_names(obj)
    if len(sensors) == 0:
        raise ValidationError("The provided multi-sensor object does not contain any data/contains no sensors.")


def is_multi_sensor_data(
    data: MultiSensorData,
    check_acc: bool = True,
    check_gyr: bool = True,
    frame: _ALLOWED_FRAMES_TYPE = "any",
    raise_exception: bool = False,
) -> bool:
    """Check if an object is a valid multi-sensor data object.

    Valid multi sensor data is:

    - is either a :class:`pandas.DataFrame` with 2 level multi-index as columns or a dictionary of single sensor
      datasets (see :func:`~gaitmap.utils.dataset_helper.is_single_sensor_data`)

    In case the data is a :class:`pandas.DataFrame` with two levels, the first level is expected to be the names
    of the used sensors.
    In both cases (dataframe or dict), `dataset[<sensor_name>]` is expected to return a valid single sensor
    dataset.
    On each of the these single-sensor datasets, :func:`~gaitmap.utils.dataset_helper.is_single_sensor_data` is used
    with the same parameters that are used to call this function.

    Parameters
    ----------
    data
        Object that should be checked
    check_acc
        If the existence of the correct acc columns should be checked
    check_gyr
        If the existence of the correct gyr columns should be checked
    frame
        The frame the dataset is expected to be in.
        This changes which columns are checked for.
        In case of "any" a dataset is considered valid if it contains the correct columns for one of the two frames.
        If you just want to check the datatype and shape, but not for specific column values, set both `check_acc` and
        `check_gyro` to `False`.
    raise_exception
        If True an exception is raised if the object does not pass the validation.
        If False, the function will return simply True or False.

    See Also
    --------
    gaitmap.utils.dataset_helper.is_single_sensor_data: Explanation and checks for single sensor data

    """
    try:
        _assert_is_dtype(data, (pd.DataFrame, dict))
        if isinstance(data, pd.DataFrame):
            _assert_has_multindex_cols(data, expected=True, nlevels=2)
        _assert_multisensor_is_not_empty(data)
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be MultiSensorData. "
                "The validation failed with the following error:\n\n{}".format(str(e))
            ) from e
        return False

    try:
        for k in get_multi_sensor_names(data):
            is_single_sensor_data(data[k], check_acc=check_acc, check_gyr=check_gyr, frame=frame, raise_exception=True)
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object appears to be MultiSensorData, "
                'but for the sensor with the name "{}", the following validation error was raised:\n\n{}'.format(
                    k, str(e)
                )
            ) from e
        return False
    return True


def is_sensor_data(
    data: SensorData, check_acc: bool = True, check_gyr: bool = True, frame: _ALLOWED_FRAMES_TYPE = "any"
) -> Literal["single", "multi"]:
    """Check if an object is valid multi-sensor or single-sensor data.

    This function will try to check the input using :func:`~gaitmap.utils.dataset_helper.is_single_sensor_data` and
    :func:`~gaitmap.utils.dataset_helper.is_multi_sensor_data`.
    In case one of the two checks is successful, a string is returned, which type of dataset the input is.
    Otherwise a descriptive error is raised

    Parameters
    ----------
    data
        Object that should be checked
    check_acc
        If the existence of the correct acc columns should be checked
    check_gyr
        If the existence of the correct gyr columns should be checked
    frame
        The frame the dataset is expected to be in.
        This changes which columns are checked for.
        In case of "any" a dataset is considered valid if it contains the correct columns for one of the two frames.
        If you just want to check the datatype and shape, but not for specific column values, set both `check_acc` and
        `check_gyro` to `False`.

    Returns
    -------
    data_type
        "single" in case of single-sensor data, "multi" in case of multi-sensor data.
        In case it is neither, a ValidationError is raised.

    See Also
    --------
    gaitmap.utils.dataset_helper.is_single_sensor_data: Explanation and checks for single sensor data
    gaitmap.utils.dataset_helper.is_multi_sensor_data: Explanation and checks for multi sensor data

    """
    try:
        is_single_sensor_data(data, check_acc=check_acc, check_gyr=check_gyr, frame=frame, raise_exception=True)
    except ValidationError as e:
        single_error = e
    else:
        return "single"

    try:
        is_multi_sensor_data(data, check_acc=check_acc, check_gyr=check_gyr, frame=frame, raise_exception=True)
    except ValidationError as e:
        multi_error = e
    else:
        return "multi"

    raise ValidationError(
        "The passed object appears to be neither single- or multi-sensor data. "
        "Below you can find the errors raised for both checks:\n\n"
        "Single-Sensor\n"
        "=============\n"
        f"{str(single_error)}\n\n"
        "Multi-Sensor\n"
        "=============\n"
        f"{str(multi_error)}"
    )


def is_multi_sensor_stride_list(
    stride_list: MultiSensorStrideList, stride_type: str, raise_exception: bool = False
) -> bool:
    """Check if an input is a multi-sensor stride list.

    A valid multi-sensor stride list is dictionary of single-sensor stride lists.

    This function :func:`~gaitmap.utils.dataset_helper.is_single_sensor_stride_list` for each of the contained stride
    lists.

    Parameters
    ----------
    stride_list
        The object that should be tested
    stride_type
        The expected stride type of this object.
        If this is "any" only the generally required columns are checked.
    raise_exception
        If True an exception is raised if the object does not pass the validation.
        If False, the function will return simply True or False.

    See Also
    --------
    gaitmap.utils.dataset_helper.is_single_sensor_stride_list: Check for multi-sensor stride lists
    gaitmap.utils.stride_list_conversion.enforce_stride_list_consistency: Remove strides that do not have the correct
        event order

    """
    try:
        _assert_is_dtype(stride_list, dict)
        _assert_multisensor_is_not_empty(stride_list)
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a MultiSensorStrideList. "
                "The validation failed with the following error:\n\n{}".format(str(e))
            ) from e
        return False

    try:
        for k in stride_list.keys():
            is_single_sensor_stride_list(stride_list[k], stride_type=stride_type, raise_exception=True)
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object appears to be a MultiSensorStrideList, "
                'but for the sensor with the name "{}", the following validation error was raised:\n\n{}'.format(
                    k, str(e)
                )
            ) from e
        return False
    return True


def _assert_has_index_columns(df: pd.DataFrame, index_cols: Iterable[Hashable]):
    ex_index_cols = list(index_cols)
    ac_index_cols = list(df.index.names)
    if ex_index_cols != ac_index_cols:
        raise ValidationError(
            "The dataframe is expected to have exactly the following index columns ({}), "
            "but it has {}".format(index_cols, df.index.name)
        )


def set_correct_index(
    df: pd.DataFrame, index_cols: Iterable[Hashable], drop_false_index_cols: bool = True
) -> pd.DataFrame:
    """Set the correct columns as index, or leave them if they are already in the index.

    Parameters
    ----------
    df
        The dataframe
    index_cols
        A list of names that correspond to the names of the multiindex level names (in order)
    drop_false_index_cols
        If True columns that are set as index in df, but shouldn't will be deleted.
        If False these columns will just be removed from the index and become regular df columns.

    Returns
    -------
    df
        A dataframe with the correct columns set as index

    """
    index_cols = list(index_cols)
    try:
        _assert_has_index_columns(df, index_cols)
        return df
    except ValidationError:
        pass

    # In case not all columns are in the the index, reset_the index and check the column names
    wrong_index = [i for i, n in enumerate(df.index.names) if n not in index_cols]
    all_wrong = len(wrong_index) == len(df.index.names)
    df_just_right_index = df.reset_index(level=wrong_index, drop=drop_false_index_cols)
    if not all_wrong:
        # In case correct indix cols are remaining make them to regular columns
        df_just_right_index = df_just_right_index.reset_index()

    try:
        _assert_has_columns(df_just_right_index, [index_cols])
    except ValidationError as e:
        raise ValidationError(
            "The dataframe is expected to have the following columns either in the index or as columns ({}), "
            "but it has {}".format(index_cols, df.columns)
        ) from e

    return df_just_right_index.set_index(index_cols)


def is_stride_list(stride_list: StrideList) -> Literal["single", "multi"]:
    """Check if an object is a valid multi-sensor or single-sensor stride list.

    This function will try to check the input using
    :func:`~gaitmap.utils.dataset_helper.is_single_sensor_stride_list` and
    :func:`~gaitmap.utils.dataset_helper.is_multi_sensor_stride_list`.
    In case one of the two checks is successful, a string is returned, which type of dataset the input is.
    Otherwise a descriptive error is raised

    Parameters
    ----------
    stride_list
        The object that should be tested
    stride_type
        The expected stride type of this object.
        If this is "any" only the generally required columns are checked.

    Returns
    -------
    dataset_type
        "single" in case of a single-sensor stride list, "multi" in case of a multi-sensor stride list.
        In case it is neither, a ValidationError is raised.

    See Also
    --------
    gaitmap.utils.dataset_helper.is_single_sensor_stride_list: Explanation and checks for single sensor stride lists
    gaitmap.utils.dataset_helper.is_multi_sensor_stride_list: Explanation and checks for multi sensor stride lists

    """
    try:
        is_single_sensor_stride_list(stride_list, stride_type="event", raise_exception=True)
    except ValidationError as e:
        single_error = e
    else:
        return "single"

    try:
        is_multi_sensor_stride_list(stride_list, stride_type="event", raise_exception=True)
    except ValidationError as e:
        multi_error = e
    else:
        return "multi"

    raise ValidationError(
        "The passed object appears to be neither a single- or a multi-sensor stride list. "
        "Below you can find the errors raised for both checks:\n\n"
        "Single-Sensor\n"
        "=============\n"
        f"{str(single_error)}\n\n"
        "Multi-Sensor\n"
        "=============\n"
        f"{str(multi_error)}"
    )


def is_single_sensor_stride_list(
    stride_list: SingleSensorStrideList, stride_type: str, raise_exception: bool = False
) -> bool:
    """Check if an input is a single-sensor stride list.

    A valid stride list:

    - is a pandas Dataframe with at least the following columns: `["s_id", "start", "end"]`.
      The `s_id` column can also be part of the index.
    - has only a single level column index
    - the value of `s_id` is unique

    Note that this function does only check the structure and not the plausibility of the contained values.
    For this `~gaitmap.utils.stride_list_conversions.enforce_stride_list_consistency` can be used.

    However, depending on the type of stride list, further requirements need to be fulfilled:

    min_vel
        A min-vel stride list describes a stride list that defines a stride from one midstance (`min_vel`) to the next.
        This type of stride list can be performed for ZUPT based trajectory estimation.
        It is expected to additionally have the following columns describing relevant stride events: `["pre_ic", "ic",
        "min_vel", "tc"]`.
        See :mod:`~gaitmap.event_detection` for details.
        For this type of stride list it is further tested, that the "start" column is actual identical to the "min_vel"
        column.
    segmented
        A segmented stride list is a stride list in which every stride starts and ends between min_vel and tc.
        For this stride list, we expect that all relevant events within each stride are already detected.
        Hence, it is checked if columns with the name `["ic", "tc", "min_vel"]` exist.
        If you want to check the structure of a stride list right after the segmentation, where no events are detected
        yet use `"any"` as `stride_type`.
    ic
        A IC stride list is a stride list in which every stride starts and ends with a IC.
        Regarding columns, it has the same requirements as the "segmented" stride list.
        Additionally it is checked, if the "start" columns is actually identical to the "ic" column.

    Parameters
    ----------
    stride_list
        The object that should be tested
    stride_type
        The expected stride type of this object.
        If this is "any" only the generally required columns are checked.
    raise_exception
        If True an exception is raised if the object does not pass the validation.
        If False, the function will return simply True or False.

    See Also
    --------
    gaitmap.utils.dataset_helper.is_multi_sensor_stride_list: Check for multi-sensor stride lists
    gaitmap.utils.stride_list_conversion.enforce_stride_list_consistency: Remove strides that do not have the correct
        event order

    """
    if stride_type == "event" or stride_type == "any":
        SL_COLS = ["ic", "tc"]
    elif stride_type == "stride":
        SL_COLS = ["start", "end", "ic", "tc"]
    else:
        raise ValueError("Stride type not of type 'event' or 'stride'.")

    try:
        _assert_is_dtype(stride_list, pd.DataFrame)
        _assert_has_multindex_cols(stride_list, expected=False)

        SL_INDEX = ["s_id"]
        stride_list = set_correct_index(stride_list, SL_INDEX)

        # Check if it has the correct columns
        all_columns = [*SL_COLS]
        _assert_has_columns(stride_list, [all_columns])

        start_event = {"min_vel": "min_vel", "ic": "ic"}
        # Check that the start time corresponds to the correct event
        if (
            start_event.get(stride_type, False)
            and len(stride_list) > 0
            and not np.array_equal(stride_list["start"].to_numpy(), stride_list[start_event[stride_type]].to_numpy())
        ):
            raise ValidationError(
                "For a {} stride list, the start column is expected to be identical to the {} column, "
                "but they are different.".format(stride_type, start_event[stride_type])
            )
        # Check that the stride ids are unique
        if not stride_list.index.nunique() == stride_list.index.size:
            raise ValidationError("The stride id of the stride list is expected to be unique.")

    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a SingleSensorStrideList. "
                "The validation failed with the following error:\n\n{}".format(str(e))
            ) from e
        return False
    return True


# ROTATIONs
def sliding_window_view(arr: np.ndarray, window_length: int, overlap: int, nan_padding: bool = False) -> np.ndarray:
    """Create a sliding window view of an input array with given window length and overlap.

    .. warning::
       This function will return by default a view onto your input array, modifying values in your result will directly
       affect your input data which might lead to unexpected behaviour! If padding is disabled (default) last window
       fraction of input may not be returned! However, if `nan_padding` is enabled, this will always return a copy
       instead of a view of your input data, independent if padding was actually performed or not!

    Parameters
    ----------
    arr : array with shape (n,) or (n, m)
        array on which sliding window action should be performed. Windowing
        will always be performed along axis 0.

    window_length : int
        length of desired window (must be smaller than array length n)

    overlap : int
        length of desired overlap (must be smaller than window_length)

    nan_padding: bool
        select if last window should be nan-padded or discarded if it not fits with input array length. If nan-padding
        is enabled the return array will always be a copy of the input array independent if padding was actually
        performed or not!

    Returns
    -------
    windowed view (or copy for nan_padding) of input array as specified, last window might be nan padded if necessary to
    match window size

    Examples
    --------
    >>> data = np.arange(0,10)
    >>> windowed_view = sliding_window_view(arr = data, window_length = 5, overlap = 3, nan_padding = True)
    >>> windowed_view
    array([[ 0.,  1.,  2.,  3.,  4.],
           [ 2.,  3.,  4.,  5.,  6.],
           [ 4.,  5.,  6.,  7.,  8.],
           [ 6.,  7.,  8.,  9., nan]])

    """
    if overlap >= window_length:
        raise ValueError("Invalid Input, overlap must be smaller than window length!")

    if window_length < 2:
        raise ValueError("Invalid Input, window_length must be larger than 1!")

    # calculate length of necessary np.nan-padding to make sure windows and overlaps exactly fits data length
    n_windows = np.ceil((len(arr) - window_length) / (window_length - overlap)).astype(int)
    pad_length = window_length + n_windows * (window_length - overlap) - len(arr)

    # had to handle 1D arrays separately
    if arr.ndim == 1:
        if nan_padding:
            # np.pad always returns a copy of the input array even if pad_length is 0!
            arr = np.pad(arr.astype(float), (0, pad_length), constant_values=np.nan)

        new_shape = (arr.size - window_length + 1, window_length)
    else:
        if nan_padding:
            # np.pad always returns a copy of the input array even if pad_length is 0!
            arr = np.pad(arr.astype(float), [(0, pad_length), (0, 0)], constant_values=np.nan)

        shape = (window_length, arr.shape[-1])
        n = np.array(arr.shape)
        o = n - shape + 1  # output shape
        new_shape = np.concatenate((o, shape), axis=0)

    # apply stride_tricks magic
    new_strides = np.concatenate((arr.strides, arr.strides), axis=0)
    view = np.lib.stride_tricks.as_strided(arr, new_shape, new_strides)[0 :: (window_length - overlap)]

    view = np.squeeze(view)  # get rid of single-dimensional entries from the shape of an array.

    return view


def _bool_fill(indices: np.ndarray, bool_values: np.ndarray, array: np.ndarray) -> np.ndarray:
    """Fill a preallocated array with bool_values.

    This method iterates over the indices and adds the values to the array at the given indices using a logical or.
    """
    for i in range(len(indices)):  # noqa: consider-using-enumerate
        index = indices[i]
        val = bool_values[i]
        index = index[~np.isnan(index)]
        # perform logical or operation to combine all overlapping window results
        array[index] = np.logical_or(array[index], val)
    return array


def find_static_samples(
    signal: np.ndarray,
    window_length: int,
    inactive_signal_th: float,
    metric: METRIC_FUNCTION_NAMES = "mean",
    overlap: int = None,
) -> np.ndarray:
    """Search for static samples within given input signal, based on windowed L2-norm thresholding.

    .. warning::
        Due to edge cases at the end of the input data where window size and overlap might not fit your data, the last
        window might be discarded for analysis and will therefore always be considered as non-static!

    Parameters
    ----------
    signal : array with shape (n, 3)
        3D signal on which static moment detection should be performed (e.g. 3D-acc or 3D-gyr data)

    window_length : int
        Length of desired window in units of samples

    inactive_signal_th : float
       Threshold to decide whether a window should be considered as active or inactive. Window will be tested on
       <= threshold

    metric : str, optional
        Metric which will be calculated per window, one of the following strings

        'mean' (default)
            Calculates mean value per window
        'maximum'
            Calculates maximum value per window
        'median'
            Calculates median value per window
        'variance'
            Calculates variance value per window

    overlap : int, optional
        Length of desired overlap in units of samples. If None (default) overlap will be window_length - 1

    Returns
    -------
    Boolean array with length n to indicate static (=True) or non-static (=False) for each sample

    Examples
    --------
    >>> test_data = load_gyro_data(path)
    >>> get_static_moments(gyro_data, window_length=128, overlap=64, inactive_signal_th = 5, metric = 'mean')

    See Also
    --------
    gaitmap.utils.array_handling.sliding_window_view: Details on the used windowing function for this method.

    """
    # test for correct input data shape
    if np.shape(signal)[-1] != 3:
        raise ValueError("Invalid signal dimensions, signal must be of shape (n,3).")
    _METRIC_FUNCTIONS = {"maximum": np.nanmax, "variance": np.nanvar, "mean": np.nanmean, "median": np.nanmedian}
    if metric not in _METRIC_FUNCTIONS:
        raise ValueError("Invalid metric passed! %s as metric is not supported." % metric)

    # check if minimum signal length matches window length
    if window_length > len(signal):
        raise ValueError(
            "Invalid window length, window must be smaller or equal than given signal length. Given signal length: "
            "%d with given window_length: %d." % (len(signal), window_length)
        )

    # add default overlap value
    if overlap is None:
        overlap = window_length - 1

    # allocate output array
    inactive_signal_bool_array = np.zeros(len(signal))

    # calculate norm of input signal (do this outside of loop to boost performance at cost of memory!)
    signal_norm = np.linalg.norm(signal, axis=1)

    mfunc = _METRIC_FUNCTIONS[metric]

    # Create windowed view of norm
    windowed_norm = np.atleast_2d(sliding_window_view(signal_norm, window_length, overlap, nan_padding=False))
    is_static = np.broadcast_to(mfunc(windowed_norm, axis=1) <= inactive_signal_th, windowed_norm.shape[::-1]).T

    # create the list of indices for sliding windows with overlap
    windowed_indices = np.atleast_2d(
        sliding_window_view(np.arange(0, len(signal)), window_length, overlap, nan_padding=False)
    )

    # iterate over sliding windows
    inactive_signal_bool_array = _bool_fill(windowed_indices, is_static, inactive_signal_bool_array)

    return inactive_signal_bool_array.astype(bool)


def _get_static_acc_vector(
    data: pd.DataFrame, window_length: int, static_signal_th: float, metric: METRIC_FUNCTION_NAMES = "median"
) -> np.ndarray:
    """Extract the mean accelerometer vector describing the static position of the sensor."""
    # find static windows within the gyro data
    if SF_GYR[0] in data.columns:
        cols = SF_GYR
    else:
        cols = SF_ACC
        print("No gyroscope data is available, acceleration signal is used for gravity alignment.")

    static_bool_array = find_static_samples(data[cols].to_numpy(), window_length, static_signal_th, metric)

    # raise exception if no static windows could be found with given user settings
    if not any(static_bool_array):
        print(
            "No static windows could be found to extract sensor offset orientation."
            "Window_length and static_singal_th will be adapted and search will be started again."
        )
        static_bool_array = find_static_samples(data[cols].to_numpy(), window_length - 5, static_signal_th + 5, metric)
        if not any(static_bool_array):
            warnings.warn(
                "No static windows could be found to extract sensor offset orientation. Please check your input data or try"
                " to adapt parameters like window_length, static_signal_th or used metric."
            )
            return None

    # get mean acc vector indicating the sensor offset orientation from gravity from static sequences
    return np.median(data[SF_ACC].to_numpy()[static_bool_array], axis=0)


def normalize(v: np.ndarray) -> np.ndarray:
    """Simply normalize a vector.

    If a 2D array is provided, each row is considered a vector, which is normalized independently.
    In case an array has norm 0, np.nan is returned.

    Parameters
    ----------
    v : array with shape (3,) or (n, 3)
         vector or array of vectors

    Returns
    -------
    normalized vector or  array of normalized vectors

    Examples
    --------
    1D array

    >>> normalize(np.array([0, 0, 2]))
    array([0., 0., 1.])

    2D array

    >>> normalize(np.array([[2, 0, 0],[2, 0, 0]]))
    array([[1., 0., 0.],
           [1., 0., 0.]])

    0 Array:

    >>> normalize(np.array([0, 0, 0]))
    array([nan, nan, nan])

    """
    v = np.array(v)
    if len(v.shape) == 1:
        ax = 0
    else:
        ax = 1
    # We do not want a warning when we divide by 0 as we expect it
    with np.errstate(divide="ignore", invalid="ignore"):
        return (v.T / np.linalg.norm(v, axis=ax)).T


def is_almost_parallel_or_antiparallel(
    v1: np.ndarray, v2: np.ndarray, rtol: float = 1.0e-5, atol: float = 1.0e-8
) -> Union[np.bool_, np.ndarray]:
    """Check if two vectors are either parallel or antiparallel.

    Parameters
    ----------
    v1 : vector with shape (3,) or array of vectors
        axis ([x, y ,z]) or array of axis
    v2 : vector with shape (3,) or array of vectors
        axis ([x, y ,z]) or array of axis
    rtol : float
        The relative tolerance parameter
    atol : float
        The absolute tolerance parameter

    Returns
    -------
    bool or array of bool values with len n

    Examples
    --------
    two vectors each of shape (3,)

    >>> is_almost_parallel_or_antiparallel(np.array([0, 0, 1]), np.array([0, 0, 1]))
    True
    >>> is_almost_parallel_or_antiparallel(np.array([0, 0, 1]), np.array([0, 1, 0]))
    False

    array of vectors

    >>> is_almost_parallel_or_antiparallel(np.array([[0, 0, 1],[0,1,0]]), np.array([[0, 0, 2],[1,0,0]]))
    array([True,False])

    """
    return np.isclose(np.abs(row_wise_dot(normalize(v1), normalize(v2))), 1, rtol=rtol, atol=atol)


def find_random_orthogonal(v: np.ndarray) -> np.ndarray:
    """Find a unitvector in the orthogonal plane to v.

    Parameters
    ----------
    v : vector with shape (3,)
         axis ([x, y ,z])

    Returns
    -------
    vector which is either crossproduct with [0,1,0] or [1,0,0].

    Examples
    --------
    two vectors each of shape (3,)

    >>> find_random_orthogonal(np.array([1, 0, 0]))
    array([0, 0, 1])

    """
    if is_almost_parallel_or_antiparallel(v, np.array([1.0, 0, 0])):
        result = np.cross(v, [0, 1, 0])
    else:
        result = np.cross(v, [1, 0, 0])
    return normalize(result)


def find_orthogonal(v1: np.ndarray, v2: np.ndarray) -> np.ndarray:
    """Return an orthogonal vector to 2 vectors.

    Parameters
    ----------
    v1 : vector with shape (3,)
         axis ([x, y ,z])
    v2 : vector with shape (3,)
         axis ([x, y ,z])

    Returns
    -------
        Returns the cross product of the two if they are not equal.

        Returns a random vector in the perpendicular plane if they are either parallel or antiparallel.
        (see :func:`find_random_orthogonal`

    Examples
    --------
    >>> find_orthogonal(np.array([1, 0, 0]),np.array([-1, 0, 0]))
    array([0, 0, -1])

    """
    if v1.ndim > 1 or v2.ndim > 1:
        raise ValueError("v1 and v2 need to be at max 1D (currently {}D and {}D".format(v1.ndim, v2.ndim))
    if is_almost_parallel_or_antiparallel(v1, v2):
        return find_random_orthogonal(v1)
    return normalize(np.cross(v1, v2))


def row_wise_dot(v1, v2, squeeze=False):
    """Calculate row wise dot product of two vectors."""
    v1, v2 = np.atleast_2d(v1, v2)
    out = np.sum(v1 * v2, axis=-1)
    if squeeze:
        return np.squeeze(out)
    return out


def find_unsigned_3d_angle(v1: np.ndarray, v2: np.ndarray) -> Union[np.ndarray, float]:
    """Find the angle (in rad) between two  3D vectors.

    Parameters
    ----------
    v1 : vector with shape (3,)  or array of vectors
        axis ([x, y ,z]) or array of axis
    v2 : vector with shape (3,) or array of vectors
        axis ([x, y ,z]) or array of axis

    Returns
    -------
        angle or array of angles between two vectors

    Examples
    --------
    two vectors: 1D

    >>> find_unsigned_3d_angle(np.array([-1, 0, 0]), np.array([-1, 0, 0]))
    0

    two vectors: 2D

    >>> find_unsigned_3d_angle(np.array([[-1, 0, 0],[-1, 0, 0]]), np.array([[-1, 0, 0],[-1, 0, 0]]))
    array([0,0])

    """
    v1_, v2_ = np.atleast_2d(v1, v2)
    v1_ = normalize(v1_)
    v2_ = normalize(v2_)
    out = np.arccos(row_wise_dot(v1_, v2_) / (np.linalg.norm(v1_, axis=-1) * np.linalg.norm(v2_, axis=-1)))
    if v1.ndim == 1:
        return np.squeeze(out)
    return out


def rotation_from_angle(axis: np.ndarray, angle: Union[float, np.ndarray]) -> Rotation:
    """Create a rotation based on a rotation axis and a angle.

    Parameters
    ----------
    axis : array with shape (3,) or (n, 3)
        normalized rotation axis ([x, y ,z]) or array of rotation axis
    angle : float or array with shape (n,)
        rotation angle or array of angeles in rad

    Returns
    -------
    rotation(s) : Rotation object with len n

    Examples
    --------
    Single rotation: 180 deg rotation around the x-axis

    >>> rot = rotation_from_angle(np.array([1, 0, 0]), np.deg2rad(180))
    >>> rot.as_quat().round(decimals=3)
    array([1., 0., 0., 0.])
    >>> rot.apply(np.array([[0, 0, 1.], [0, 1, 0.]])).round()
    array([[ 0., -0., -1.],
           [ 0., -1.,  0.]])

    Multiple rotations: 90 and 180 deg rotation around the x-axis

    >>> rot = rotation_from_angle(np.array([1, 0, 0]), np.deg2rad([90, 180]))
    >>> rot.as_quat().round(decimals=3)
    array([[0.707, 0.   , 0.   , 0.707],
           [1.   , 0.   , 0.   , 0.   ]])
    >>> # In case of multiple rotations, the first rotation is applied to the first vector
    >>> # and the second to the second
    >>> rot.apply(np.array([[0, 0, 1.], [0, 1, 0.]])).round()
    array([[ 0., -1.,  0.],
           [ 0., -1.,  0.]])

    """
    angle = np.atleast_2d(angle)
    axis = np.atleast_2d(axis)
    return Rotation.from_rotvec(np.squeeze(axis * angle.T))


def find_shortest_rotation(v1: np.ndarray, v2: np.ndarray) -> Rotation:
    """Find a quaternion that rotates v1 into v2 via the shortest way.

    Parameters
    ----------
    v1 : vector with shape (3,)
        axis ([x, y ,z])
    v2 : vector with shape (3,)
        axis ([x, y ,z])

    Returns
    -------
    rotation
        Shortest rotation that rotates v1 into v2

    Examples
    --------
    >>> goal = np.array([0, 0, 1])
    >>> start = np.array([1, 0, 0])
    >>> rot = find_shortest_rotation(start, goal)
    >>> rotated = rot.apply(start)
    >>> rotated
    array([0., 0., 1.])

    """
    if (not np.isclose(np.linalg.norm(v1, axis=-1), 1)) or (not np.isclose(np.linalg.norm(v2, axis=-1), 1)):
        raise ValueError("v1 and v2 must be normalized")
    axis = find_orthogonal(v1, v2)
    angle = find_unsigned_3d_angle(v1, v2)
    return rotation_from_angle(axis, angle)


def get_gravity_rotation(gravity_vector: np.ndarray, expected_gravity: np.ndarray = GRAV_VEC) -> Rotation:
    """Find the rotation matrix needed to align  z-axis with gravity.

    Parameters
    ----------
    gravity_vector : vector with shape (3,)
        axis ([x, y ,z])
    expected_gravity : vector with shape (3,)
        axis ([x, y ,z])

    Returns
    -------
    rotation
        rotation between given gravity vector and the expected gravity

    Examples
    --------
    >>> goal = np.array([0, 0, 1])
    >>> start = np.array([1, 0, 0])
    >>> rot = get_gravity_rotation(start)
    >>> rotated = rot.apply(start)
    >>> rotated
    array([0., 0., 1.])

    """
    gravity_vector = normalize(gravity_vector)
    expected_gravity = normalize(expected_gravity)
    return find_shortest_rotation(gravity_vector, expected_gravity)


def align_dataset_to_gravity(
    dataset: SensorData,
    sampling_rate_hz: float,
    window_length_s: float = 0.7,
    static_signal_th: float = 2.5,
    metric: METRIC_FUNCTION_NAMES = "median",
    gravity: np.ndarray = GRAV_VEC,
) -> SensorData:
    """Align dataset, so that each sensor z-axis (if multiple present in dataset) will be parallel to gravity.

    Median accelerometer vector will be extracted form static windows which will be classified by a sliding window
    with (window_length -1) overlap and a thresholding of the gyro signal norm. This will be performed for each sensor
    in the dataset individually.

    Parameters
    ----------
    dataset : gaitmap.utils.dataset_helper.Sensordata
        dataframe representing a single or multiple sensors.
        In case of multiple sensors a df with MultiIndex columns is expected where the first level is the sensor name
        and the second level the axis names (all sensor frame axis must be present)

    sampling_rate_hz: float
        Samplingrate of input signal in units of hertz.

    window_length_s : float
        Length of desired window in units of seconds.

    static_signal_th : float
       Threshold to decide whether a window should be considered as static or active. Window will be classified on
       <= threshold on gyro norm

    metric
        Metric which will be calculated per window, one of the following strings

        'maximum' (default)
            Calculates maximum value per window
        'mean'
            Calculates mean value per window
        'median'
            Calculates median value per window
        'variance'
            Calculates variance value per window

    gravity : vector with shape (3,), axis ([x, y ,z]), optional
        Expected direction of gravity during rest after the rotation.
        For example if this is `[0, 0, 1]` the sensor will measure +g on the z-axis after rotation (z-axis pointing
        upwards)

    Returns
    -------
    aligned dataset
        This will always be a copy. The original dataframe will not be modified.

    Examples
    --------
    >>> # pd.DataFrame containing one or multiple sensor data streams, each of containing all 6 IMU
    ... # axis (acc_x, ..., gyr_z)
    >>> dataset_df = ...
    >>> align_dataset_to_gravity(dataset_df, window_length_s = 0.7, static_signal_th = 2.0, metric = 'median',
    ... gravity = np.array([0.0, 0.0, 1.0])
    <copy of dataset with all axis aligned to gravity>

    See Also
    --------
    gaitmap.utils.static_moment_detection.find_static_sequences: Details on the used static moment detection function
        for this method.

    """
    dataset_type = is_sensor_data(dataset, check_gyr=False)

    window_length = int(round(window_length_s * sampling_rate_hz))
    acc_vector: Union[np.ndarray, Dict[Hashable, np.ndarray]]
    if dataset_type == "single":
        # get static acc vector
        acc_vector = _get_static_acc_vector(dataset, window_length, static_signal_th, metric)
        if acc_vector is not None:
            # get rotation to gravity
            rotation = get_gravity_rotation(acc_vector, gravity)
        else:
            warnings.warn("Align to gravity was ignored as not static vector is available.")
            return dataset
    else:
        # build dict with static acc vectors for each sensor in dataset
        acc_vector = {
            name: _get_static_acc_vector(dataset[name], window_length, static_signal_th, metric)
            for name in get_multi_sensor_names(dataset)
        }
        # build rotation dict for each dataset from acc dict and gravity
        if any([v is None for v in acc_vector.values()]):
            warnings.warn("Align to gravity was ignored as not static vector is available.")
            return dataset
        rotation = {name: get_gravity_rotation(acc_vector[name], gravity) for name in get_multi_sensor_names(dataset)}
    return rotate_dataset(dataset, rotation)


def _rotate_sensor(
    data: SingleSensorData, rotation: Optional[Rotation], inplace: bool = False, gyr_avail: bool = True
) -> SingleSensorData:
    """Rotate the data of a single sensor with acc and gyro."""
    if inplace is False:
        data = data.copy()
    if rotation is None:
        return data
    if gyr_avail:
        data[SF_GYR] = rotation.apply(data[SF_GYR].to_numpy())
    data[SF_ACC] = rotation.apply(data[SF_ACC].to_numpy())
    return data


def rotate_dataset(
    dataset: SensorData, rotation: Union[Rotation, Dict[str, Rotation]], check_gyr: bool = True
) -> SensorData:
    """Apply a rotation to acc and gyro data of a dataset.

    Parameters
    ----------
    dataset
        dataframe representing a single or multiple sensors.
        In case of multiple sensors a df with MultiIndex columns is expected where the first level is the sensor name
        and the second level the axis names (all sensor frame axis must be present)
    rotation
        In case a single rotation object is passed, it will be applied to all sensors of the dataset.
        If a dictionary of rotations is applied, the respective rotations will be matched to the sensors based on the
        dict keys.
        If no rotation is provided for a sensor, it will not be modified.

    Returns
    -------
    rotated dataset
        This will always be a copy. The original dataframe will not be modified.

    Examples
    --------
    This will apply the same rotation to the left and the right foot

    >>> dataset = ...  # Sensordata with a left and a right foot sensor
    >>> rotate_dataset(dataset, rotation=rotation_from_angle(np.array([0, 0, 1]), np.pi))
    <copy of dataset with all axis rotated>

    This will apply different rotations to the left and the right foot

    >>> dataset = ...  # Sensordata with a left and a right foot sensor (sensors called "left" and "right")
    >>> rotate_dataset(dataset, rotation={'left': rotation_from_angle(np.array([0, 0, 1]), np.pi),
    ...     'right':rotation_from_angle(np.array([0, 0, 1]), np.pi / 2))
    <copy of dataset with all axis rotated>

    See Also
    --------
    gaitmap.utils.rotations.rotate_dataset_series: Apply a series of rotations to a dataset

    """
    dataset_type = is_sensor_data(dataset, frame="sensor", check_gyr=check_gyr)
    if dataset_type == "single":
        if isinstance(rotation, dict):
            raise ValueError(
                "A Dictionary for the `rotation` parameter is only supported if a MultiIndex dataset (named sensors) is"
                " passed."
            )
        return _rotate_sensor(dataset, rotation, inplace=False, gyr_avail=check_gyr)

    rotation_dict = rotation
    if not isinstance(rotation_dict, dict):
        rotation_dict = {k: rotation for k in get_multi_sensor_names(dataset)}

    if isinstance(dataset, dict):
        rotated_dataset = {**dataset}
        original_cols = None
    else:
        rotated_dataset = dataset.copy()
        original_cols = dataset.columns
    for key in rotation_dict.keys():
        test = _rotate_sensor(dataset[key], rotation_dict[key], inplace=False, gyr_avail=check_gyr)
        rotated_dataset[key] = test

    if isinstance(dataset, pd.DataFrame):
        # Restore original order
        rotated_dataset = rotated_dataset[original_cols]
    return rotated_dataset


def convert_left_foot_to_fbf(data: SingleSensorData):
    """Convert the axes from the left foot sensor frame to the foot body frame (FBF).

    This function assumes that your dataset is already aligned to the gaitmap FSF.

    Parameters
    ----------
    data
        raw data frame containing acc and gyr data

    Returns
    -------
    converted data frame

    See Also
    --------
    gaitmap.utils.coordinate_conversion.convert_right_foot_to_fbf: conversion of right foot SingleSensorData
    gaitmap.utils.coordinate_conversion.convert_to_fbf: convert multiple sensors at the same time

    """
    is_single_sensor_data(data, check_gyr=False, frame="sensor", raise_exception=True)

    if "gyr_x" in data.columns:
        cols_bf = BF_COLS
    else:
        cols_bf = BF_ACC

    result = pd.DataFrame(columns=cols_bf)

    # Loop over all axes and convert each one separately
    for sf_col_name in data.columns:
        result[FSF_FBF_CONVERSION_LEFT[sf_col_name][1]] = FSF_FBF_CONVERSION_LEFT[sf_col_name][0] * data[sf_col_name]

    return result


def convert_right_foot_to_fbf(data: SingleSensorData):
    """Convert the axes from the right foot sensor frame to the foot body frame (FBF).

    This function assumes that your dataset is already aligned to the gaitmap FSF.

    Parameters
    ----------
    data
        raw data frame containing acc and gyr data

    Returns
    -------
    converted data frame

    See Also
    --------
    gaitmap.utils.coordinate_conversion.convert_left_foot_to_fbf: conversion of left foot SingleSensorData
    gaitmap.utils.coordinate_conversion.convert_to_fbf: convert multiple sensors at the same time

    """
    is_single_sensor_data(data, check_acc=False, frame="sensor", raise_exception=True)

    result = pd.DataFrame(columns=BF_COLS)

    # Loop over all axes and convert each one separately
    for sf_col_name in SF_COLS:
        result[FSF_FBF_CONVERSION_RIGHT[sf_col_name][1]] = FSF_FBF_CONVERSION_RIGHT[sf_col_name][0] * data[sf_col_name]

    return result


def _handle_foot(foot, foot_like, data, rot_func):
    result = dict()
    if foot_like:
        foot = [sensor for sensor in get_multi_sensor_names(data) if foot_like in sensor]
        if not foot:
            warnings.warn(
                "The substring {} is not contained in any sensor name. Available sensor names are: {}".format(
                    foot_like, get_multi_sensor_names(data)
                )
            )
    foot = foot or []
    for s in foot:
        if s not in data:
            raise KeyError("Sensordata contains no sensor with name " + s)
        result[s] = rot_func(data[s])
    return result


def convert_to_fbf(
    data: MultiSensorData,
    left: Optional[List[str]] = None,
    right: Optional[List[str]] = None,
    right_like: str = None,
    left_like: str = None,
):
    """Convert the axes from the sensor frame to the body frame for one MultiSensorDataset.

    This function assumes that your dataset is already aligned to the gaitmap FSF.
    Sensors that should not be transformed are kept untouched.
    Note, that the column names of all transformed dataset is changed to the respective body frame names.

    This function can handle multiple left and right sensors at the same time.

    Parameters
    ----------
    data
        MultiSensorDataset
    left
        List of strings indicating sensor names which will be rotated using definition of left conversion.
        This option can not be used in combination with `left_like`.
    right
        List of strings indicating sensor names which will be rotated using definition of right conversion
        This option can not be used in combination with `right_like`.
    left_like
        Consider all sensors containing this string in the name as left foot sensors.
        This option can not be used in combination with `left`.
    right_like
        Consider all sensors containing this string in the name as right foot sensors.
        This option can not be used in combination with `right`.


    Returns
    -------
    converted MultiSensorDataset

    Examples
    --------
    These examples assume that your dataset has two sensors called `left_sensor` and `right_sensor`.

    >>> dataset = ... # Sensordata in FSF
    >>> fbf_dataset = convert_to_fbf(dataset, left_like="left_", right_like="right_")

    Alternatively, you can specify the full sensor names.

    >>> fbf_dataset = convert_to_fbf(dataset, left=["left_sensor"], right_sensor=["right_sensor"])

    See Also
    --------
    gaitmap.utils.coordinate_conversion.convert_left_foot_to_fbf: conversion of left foot SingleSensorData
    gaitmap.utils.coordinate_conversion.convert_right_foot_to_fbf: conversion of right foot SingleSensorData

    """
    if not is_multi_sensor_data(data, frame="sensor", check_gyr=False):
        raise ValueError("No valid FSF MultiSensorDataset supplied.")

    if (left and left_like) or (right and right_like) or not any((left, left_like, right, right_like)):
        raise ValueError(
            "You need to either supply a list of names via the `left` or `right` arguments, or a single string for the "
            "`left_like` or `right_like` arguments, but not both!"
        )

    left_foot = _handle_foot(left, left_like, data, rot_func=convert_left_foot_to_fbf)
    right_foot = _handle_foot(right, right_like, data, rot_func=convert_right_foot_to_fbf)

    sensor_names = get_multi_sensor_names(data)
    result = {k: data[k] for k in sensor_names}
    result = {**result, **left_foot, **right_foot}

    # If original data is not synchronized (dictionary), return as dictionary
    if isinstance(data, dict):
        return result
    # For synchronized sensors, return as MultiIndex dataframe
    df = pd.concat(result, axis=1)
    # restore original order
    return df[sensor_names]



"""Set of helper functions."""
import pickle
from pathlib import Path

import pandas as pd
from scipy import signal

from eargait.utils.helper_gaitmap import SensorData, is_sensor_data


def load_pickle(path: Path):
    """Load a pickle file."""
    with open(path, "rb") as handle:
        b = pickle.load(handle)
    return b


def save_pickle(path: Path, variable):
    """Save a variable as pickle."""
    path.parent.mkdir(parents=True, exist_ok=True)
    with open(path, "wb") as handle:
        pickle.dump(variable, handle, protocol=pickle.HIGHEST_PROTOCOL)


def butter_lowpass_filter(data: SensorData, cutoff_freq: int, nyq_freq: int, order: int) -> SensorData:
    """Butterworth Low Pass Filter."""
    is_sensor_data(data)
    normal_cutoff = float(cutoff_freq) / nyq_freq
    if normal_cutoff == 1:
        normal_cutoff = 10 / nyq_freq
    b, a = signal.butter(order, normal_cutoff, btype="lowpass")
    return _butter_filter(data, a, b)


def butter_highpass_filter(data: SensorData, cutoff_freq: int, nyq_freq: int, order: int) -> SensorData:
    """Butterworth High Pass Filter."""
    is_sensor_data(data)
    normal_cutoff = float(cutoff_freq) / nyq_freq
    b, a = signal.butter(order, normal_cutoff, btype="highpass")
    return _butter_filter(data, a, b)


def _butter_filter(data: SensorData, a, b) -> SensorData:
    kind = is_sensor_data(data)
    if kind == "single":
        data_filtered = _single_sensor_butter_filter(data, a, b)
    elif kind == "multi":
        data_filtered = {}
        for key, val in data.items():
            data_filtered[key] = _single_sensor_butter_filter(val, a, b)
    return data_filtered


def _single_sensor_butter_filter(data: pd.DataFrame, a, b) -> pd.DataFrame:
    assert isinstance(data, pd.DataFrame)
    data_filt = data.copy()
    for col in data.columns:
        try:
            data_filt[col] = signal.filtfilt(b, a, data[col])
        except ValueError:
            data_filt[col] = signal.filtfilt(b, a, data[col], padlen=len(data) - 5)
    return data_filt



"""Class for all event detection algorithms based on ear worn motion sensors."""
from typing import Dict, TypeVar, Union

import pandas as pd

from eargait.base.base_eargait import BaseEarGait
from eargait.utils.gait_parameters import (
    combine_spatial_temporal,
    get_average_spatial_params,
    get_average_spatiotemp_params,
    get_average_temporal_params,
    get_temporal_params,
)
from eargait.utils.helper_gaitmap import Hashable, SensorData, get_multi_sensor_names, is_sensor_data, is_stride_list
from eargait.utils.helpers import butter_lowpass_filter

Self = TypeVar("Self", bound="EarGait")


class EarGait(BaseEarGait):
    """Base class for gait analysis using ear worn sensors.

    Parameters
    ----------
    data
        The IMU data passed to the `detect` method.
    sample_rate_hz
        The sample rate of the data


    Attributes
    ----------
    event_detection_method
        Method used to detect gait events, typically DiaoAdaptedEventDetection
    event_list
        Event list or dictionary with such values.
        The result of the `detect` method holding all temporal gait events. A stride is defined
        such that terminate contact is follwed by an initial contact of the same foot.
    event_list_consistent_
        Event list with consistent gait events, hence, an alternating sequence of ipsi- and contralateral strides.
        Non consistent (non-alternating) step are set to NaN.
    temporal_params
        Temporal parameter of the strides in a gait sequence given the event_list_.
    average_temporal_params
        Average temporal parameters of a gait sequence.
    spatial_params
        Spatial parameter of the strides in a gait sequence.
    average_spatial_params
        Average spatial parameters of a gait sequence.
    spatiotemporal_params
        Spatiotemporal parameter of the strides in a gait sequence.
    average_spatiotemporal_params
        Average spatiotemporal parameters of a gait sequence.
    step_counter
        Number of steps in a gait sequences. Calculated by the sum over the number if initial contacts in the
        event_list_.
    cadence
        Frequency of steps averaged over a gait sequence in steps/minute.


    Examples
    --------
    Gait Analysis for ear-worn sensors
    >>> gait_event_detection_method = DiaoAdaptedEventDetection(200,200)
    >>> eargait = EarGait(200, gait_event_detection_method, True)
    >>> eargait = detect(data)
    >>> average_temporal_params = eargait.average_temporal_params()

    ...

    Notes
    -----
    terminal contact (`tc`), often also referred to as toe off:
        At `tc` the movement of the ankle joint changes from a plantar flexion to a dorsal extension in the sagittal
        plane..

    initial contact (`ic`), often also referred to as heel strike:
        At `ic` the foot decelerates rapidly when the foot hits the ground.

    ipsilateral and contralateral:
        In contrast to foot worn sensors, ear worn sensors contain gait events from left and right foot.
        A stride of the foot situated at same body side as the ear worn sensor is referred to as ipsilateral stride.
        A stride of the foot situated at the oppisite body side if referred to as contralateral stride.
        A gait sequence is a steady alternating sequence of ipsi- and contralateral strides.

    """

    sample_rate_hz: float
    event_detection_method = None
    spatial_params_method = None
    bool_use_event_list_consistent: True

    data: SensorData = None
    event_list: Union[pd.DataFrame, Dict[str, pd.DataFrame]] = None

    _temporal_params_memory = None
    _spatial_params_memory = None
    _spatiotemporal_params_memory = None

    def __init__(
        self,
        sample_rate_hz: float,
        event_detection_method=None,
        spatial_params_method=None,
        bool_use_event_list_consistent: bool = False,
    ):
        super().__init__()
        self.sample_rate_hz = sample_rate_hz
        self.event_detection_method = event_detection_method
        if isinstance(spatial_params_method, bool):
            raise ValueError(
                "Input for spatial method can not be of type bool. "
                "Please provide spatial parameter estimation method or None."
            )
        self.spatial_params_method = spatial_params_method
        self.bool_use_event_list_consistent = bool_use_event_list_consistent

    def detect(self: Self, data: SensorData) -> Self:  # noqa
        """Find gait events in data within strides provided by stride_list.

        Parameters
        ----------
        data
            The data set holding the imu raw data

        sample_rate_hz
            The sample rate of the data

        Returns
        -------
        self
            The class instance with all result attributes populated

        """
        self._clear_all_variable()
        self.data = data
        if self.event_detection_method is None:
            raise ValueError(
                "No event detection method was specified. Please define a event detection method, "
                "e.g. DiaoAdaptedEventDetection"
            )
        self.event_detection_method.detect(self.data)
        if not self.bool_use_event_list_consistent:
            self.event_list = self.event_detection_method.event_list_
        else:
            self.event_list = self.event_detection_method.event_list_consistent_
        return self

    @property
    def temporal_params(self) -> pd.DataFrame:
        """Return temporal gait parameters."""
        if self._temporal_params_memory is None:
            assert self._has_event_list()
            temporal_params = get_temporal_params(self.event_list, self.sample_rate_hz)
            self._temporal_params_memory = temporal_params
        return self._temporal_params_memory

    @property
    def spatial_params(self) -> pd.DataFrame:
        """Return spatial gait parameters."""
        if self._spatial_params_memory is None:
            assert self._has_event_list()

            if is_stride_list(self.event_list) == "single":
                spatial_params = self.spatial_params_method.estimate(self.data, self.event_list)
            else:
                spatial_params = {}
                for sensor, events in self.event_list.items():
                    spatial_params[sensor] = self.spatial_params_method.estimate(self.data[sensor], events)
            self._spatial_params_memory = spatial_params
        return self._spatial_params_memory

    @property
    def average_temporal_params(self) -> pd.DataFrame:
        """Return average temporal gait parameters."""
        average_temporal_params = get_average_temporal_params(self.temporal_params)
        return average_temporal_params

    @property
    def average_spatial_params(self) -> pd.DataFrame:
        """Return average spatial gait parameters."""
        average_spatial_params = get_average_spatial_params(self.spatial_params)
        return average_spatial_params

    @property
    def spatiotemporal_params(self) -> pd.DataFrame:
        """Return spatiotemporal gait parameters."""
        if self._spatiotemporal_params_memory is None:
            temp = self.temporal_params
            spat = self.spatial_params
            assert len(temp) == len(spat)
            spatiotemporal = combine_spatial_temporal(temp, spat)
            self._spatiotemporal_params_memory = spatiotemporal
        return self._spatiotemporal_params_memory

    @property
    def average_spatiotemporal_params(self) -> pd.DataFrame:
        """Return average spatiotemporal gait parameters."""
        average_spatiotemp = get_average_spatiotemp_params(self.spatiotemporal_params)
        return average_spatiotemp

    @property
    def number_of_steps(self: Self) -> int:
        """Get number of steps of a gait sequence."""
        assert self._has_event_list()
        event_list_type = is_stride_list(self.event_list)
        events_list = self.event_list
        if event_list_type == "single":
            results = self._single_step_counter(events_list)
        else:
            results: Dict[Hashable, Dict[str, pd.DataFrame]] = {}
            for sensor in get_multi_sensor_names(events_list):
                results[sensor] = self._single_step_counter(events_list[sensor])
        return results

    @property
    def cadence(self: Self) -> float:
        """Get the cadence of a gait sequence based on number of step divided by duration."""
        steps = self.number_of_steps
        if is_stride_list(self.event_list) == "single":
            results = self._single_cadence(self.event_list, steps, self.sample_rate_hz)
        else:
            results: Dict[Hashable, Dict[str, pd.DataFrame]] = {}
            for sensor in get_multi_sensor_names(self.event_list):
                results[sensor] = self._single_cadence(self.event_list[sensor], steps[sensor], self.sample_rate_hz)
        return results

    @property
    def cadence_dominant_freq(self: Self) -> float:
        """Get the cadence of a gait sequence based on dominant frequency of a walking bout."""
        if self.data is not None:
            dataset_type = is_sensor_data(self.data, frame="body")
            filtered_data = butter_lowpass_filter(self.data, 20, self.sample_rate_hz / 2, 4)
            if dataset_type == "single":
                results = self._single_cadence_dominant_freq(filtered_data, self.sample_rate_hz)
            else:
                results: Dict[Hashable, Dict[str, pd.DataFrame]] = {}
                for sensor in get_multi_sensor_names(self.data):
                    results[sensor] = self._single_cadence_dominant_freq(filtered_data[sensor], self.sample_rate_hz)
            return results
        return None

    def get_gait_parameters(self):
        """Get the average gait parameters of a gait sequence."""
        assert self._has_event_list()
        average_spatiotemp = self.average_spatiotemporal_params
        asy = self.get_asymmetry()
        si = self.get_symmetry_index()
        vari = self.get_variability()
        no_steps = self.number_of_steps
        cad = self.cadence
        cad_harm = self.cadence_dominant_freq

        if is_stride_list(self.event_list) == "single":
            gait_params = self.single_gait_parameters(average_spatiotemp, asy, si, vari, no_steps, cad, cad_harm)
        else:
            gait_params = {}
            for se in self.event_list.keys():
                df = self.single_gait_parameters(
                    average_spatiotemp[se], asy[se], si[se], vari[se], no_steps[se], cad[se], cad_harm[se]
                )
                gait_params[se] = df
        return gait_params

    @staticmethod
    def single_gait_parameters(average_spatiotemp, asy, si, vari, no_steps, cad, cad_harm):
        df = pd.Series(dtype="float64")
        df = pd.concat([df, average_spatiotemp.loc["mean"]])
        df = pd.concat([df, asy, si])
        df = pd.concat([df, vari])
        df = pd.concat([df, pd.Series(no_steps, index=["number_of_steps"])])
        df = pd.concat([df, pd.Series(cad, index=["cadence"])])
        df = pd.concat([df, pd.Series(cad_harm, index=["cadence_dom_freq"])])
        velocity_harm = cad_harm * df.step_length
        velocity = cad * df.step_length
        df = pd.concat([df, pd.Series([velocity, velocity_harm], index=["gait_velocity", "gait_velocity_dom_freq"])])
        return df

    def plot(self: Self, plot_ssa: bool = False):
        """Plot gait data and events."""
        assert self._has_event_list()
        if self.event_detection_method is not None:
            self.event_detection_method.plot(plot_ssa)
        else:
            self._plot()

    def get_asymmetry(self):
        """Get asymmetry for temporal and spatial parameters.

        Absolute and in percent.
        """
        if is_stride_list(self.event_list) == "single":
            asymmetry = self._single_asymmetry(self.spatiotemporal_params, self.average_spatiotemporal_params)
        else:
            asymmetry = {}
            for sensor, temps in self.spatiotemporal_params.items():
                asymmetry[sensor] = self._single_asymmetry(temps, self.average_spatiotemporal_params[sensor])
        return asymmetry

    def get_symmetry_index(self):
        """Get symmetry index for temporal and spatial parameters.

        Defined by:
        [Zhao et al., IMU-based Gait Analysis for Rehabilitation Assessment of Patients with Gait Disorders, IEEE 2017]

        """
        assert self._has_event_list()
        if is_stride_list(self.event_list) == "single":
            symmetry = self._single_symmetry_index(self.spatiotemporal_params)
        else:
            symmetry = {}
            for sensor, temps in self.spatiotemporal_params.items():
                symmetry[sensor] = self._single_symmetry_index(temps)
        return symmetry

    def get_variability(self):
        """Get variability for spatiotemporal parameters.

        STD: standard deviation
        CV: Coefficient Variation: ration STD/MEAN

        """
        assert self._has_event_list()

        if is_stride_list(self.event_list) == "single":
            variability = self._single_variability(self.average_spatiotemporal_params)
        else:
            variability = {}
            for sensor, temps in self.average_spatiotemporal_params.items():
                variability[sensor] = self._single_variability(temps)
        return variability



"""A set of functions that help to load and rotate the dataset."""
from eargait.preprocessing.load_data_helpers import load
from eargait.preprocessing.rotations import aling_gravity_and_convert_ear_to_ebf, convert_ear_to_ebf

__all__ = ["load", "convert_ear_to_ebf", "aling_gravity_and_convert_ear_to_ebf"]



"""Base class for all event detection algorithms based on ear worn motion sensors."""
import warnings
from typing import TypeVar

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from numpy.linalg import norm

from eargait.utils.helper_gaitmap import BF_ACC, BF_GYR, is_sensor_data

Self = TypeVar("Self", bound="BaseEarGait")


class BaseEarGait:
    """Base Class for EarGait."""

    def __init__(self):
        self._temporal_params_memory = None
        self._spatial_params_memory = None
        self._spatiotemporal_params_memory = None

    def _plot(self):
        assert self.data is not None
        dataset_type = is_sensor_data(self.data, frame="body")
        if dataset_type == "single":
            self._plot_single(self.data)
        else:
            for sensor in self.data.keys():
                self._plot_single(self.data[sensor])

    def _has_event_list(self: Self):
        if not hasattr(self, "event_list"):
            warnings.warn("No event_list exists. Please run self.detect().")
            return False
        return True

    @staticmethod
    def _plot_single(data):
        fau_blaues = ["#003865", "#00b1eb", "#98a4ae"]
        _, axes = plt.subplots(2, 1)
        for col, c in zip(BF_ACC, fau_blaues):
            axes[0].plot(data[col], label=col, c=c)
        axes[0].legend()
        for col, c in zip(BF_GYR, fau_blaues):
            axes[1].plot(data[col], label=col, c=c)
        axes[1].legend()

    @staticmethod
    def _single_step_counter(event_list):
        return event_list.ic.shape[0]

    @staticmethod
    def _single_cadence(event_list, no_steps, sample_rate) -> int:
        ics = event_list.ic.to_numpy()
        duration = (ics[-1] - ics[0]) / sample_rate
        return (no_steps - 0) / duration

    @staticmethod
    def _single_cadence_dominant_freq(filtered_data, sample_rate_hz):
        signal_1d = norm(filtered_data[BF_ACC], axis=1)
        signal_1d = np.reshape(signal_1d, (1, signal_1d.shape[0]))
        locomotion_band = (0.3, 4)
        lower_bound = int(np.floor(sample_rate_hz / locomotion_band[1]))
        # print("lower bound:", lower_bound)
        # (sampling_rate / locomotion_band_lower = upper bound of autocorrelation)
        upper_bound = int(np.ceil(sample_rate_hz / locomotion_band[0]))
        # print("upper bound:", upper_bound)
        # autocorr from 0-upper motion band
        auto_corr = np.empty((signal_1d.shape[0], upper_bound + 1))
        for tau in range(upper_bound + 1):
            tmax = signal_1d.shape[1] - tau
            umax = signal_1d.shape[1] + tau
            auto_corr[:, tau] = (signal_1d[:, :tmax] * signal_1d[:, tau:umax]).sum(axis=1)
        # calculate dominant frequency in Hz
        dominant_frequency = (
            1 / (np.argmax(auto_corr[:, lower_bound:], axis=-1) + lower_bound).astype(float) * sample_rate_hz
        )
        return dominant_frequency[0]

    @staticmethod
    def _single_consistent_event_list(event_list):
        tmp = event_list.copy()
        tmp["diff"] = 1
        tmp.loc[tmp.side == "contralateral", "diff"] = -1
        tmp["diff"] = tmp["diff"].diff()
        r = np.abs(np.diff(np.sign(tmp["diff"]))) == 2
        r[0] = True
        # counter = sum(~(r))
        r = np.append(r, True)
        event_list.loc[~r, ["ic", "tc"]] = np.nan
        return event_list

    @staticmethod
    def _single_symmetry_index(spatiotemp_params):
        ipsi = spatiotemp_params.loc[spatiotemp_params.side == "ipsilateral"].mean(numeric_only=True)
        contra = spatiotemp_params.loc[spatiotemp_params.side == "contralateral"].mean(numeric_only=True)
        si = 2 * np.abs((ipsi - contra) / (ipsi + contra)) * 100
        si = si.add_suffix("_si")
        return si

    @staticmethod
    def _single_variability(spatiotemp_params):
        varia = spatiotemp_params.loc["std"].add_suffix("_std")
        varia = pd.concat([varia, (spatiotemp_params.loc["std"] / spatiotemp_params.loc["mean"]).add_suffix("_cv")])
        return varia

    @staticmethod
    def _single_asymmetry(spatiotemp_params, average_spatiotemp_params):
        ipsi = spatiotemp_params.loc[spatiotemp_params.side == "ipsilateral"].mean(numeric_only=True)
        contra = spatiotemp_params.loc[spatiotemp_params.side == "contralateral"].mean(numeric_only=True)
        diff = np.abs(ipsi - contra)
        diff_percent = diff / average_spatiotemp_params.loc["mean"]
        diff = diff.add_suffix("_asymmetry")
        diff_percent = diff_percent.add_suffix("_asymmetry_percent")
        diff = pd.concat([diff, diff_percent])
        return diff

    def _has_subject_data(self):
        if any([self.age, self.height, self.gender]):
            raise ValueError(
                "Subjects characteristic are not given: age, gender height. Please set by using self.set_subject_data"
            )

    def _clear_all_variable(self):
        self._temporal_params_memory = None
        self._spatial_params_memory = None
        self._spatiotemporal_params_memory = None



"""Event detection algorithm by Jarchi. Not recommended to use."""
import warnings
from typing import Dict, Tuple, TypeVar, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from pyts.decomposition import SingularSpectrumAnalysis
from scipy.signal import argrelextrema, find_peaks, firwin, kaiserord, lfilter  # noqa

from eargait.event_detection.base_event_detection import BaseEventDetection
from eargait.utils.consts import BF_ACC
from eargait.utils.helper_gaitmap import SensorData

Self = TypeVar("Self", bound="JarchiEventDetection")

# pylint: skip-file


class JarchiEventDetection(BaseEventDetection):
    """Find gait events in the IMU raw signal based on signal characteristics.

    RamppEventDetection uses signal processing .... [cite]

    Parameters
    ----------
    window_length
        The window length for the SSA.

    Attributes
    ----------
    event_list_ : ?? A stride list or dictionary with such values
        The result of the `detect` method holding all temporal gait events and start / end of all strides.
        This version of the results has the same stride borders than the input `stride_list` and has additional columns
        for all the detected events.
        Strides for which no valid events could be found are removed.

    walking_speed_

    Other Parameters
    ----------------
    data
        The data passed to the `detect` method.
    sampling_rate_hz
        The sampling rate of the data


    Examples
    --------
    Get gait events from single sensor signal

    >>> event_detection = JarchiEventDetection()
    >>> event_detection.detect(data=data, sampling_rate_hz=200.0)
            ic      tc       side
    s_id
    0      651.0    584.0    ipsilateral
    1      839.0    802.0    contralateral
    2      1089.0   1023.0   ipsilateral
    ...

    Notes
    -----
    terminal contact (`tc`), originally called toe-off (TO):

    initial contact (`ic`), originally called heel strike (HS):

    .. [1] Jarchi et al...

    """

    data: SensorData
    sampling_rate_hz: float

    window_length: int
    ssa: SingularSpectrumAnalysis
    t1: int

    event_list_: Union[pd.DataFrame, Dict[str, pd.DataFrame]]

    def __init__(
        self,
        window_length: int = 200,
        sampling_rate_hz: int = 200,
    ):
        # raise ValueError("Jarchi event detection method is not fully implemented and hence not recommended to use.")
        message = "Jarchi event detection method is not fully implemented and hence not recommended to use."
        warnings.warn(message)

        super(JarchiEventDetection, self).__init__()  # noqa
        self.window_length = window_length
        self.sampling_rate_hz = sampling_rate_hz
        self.ssa = SingularSpectrumAnalysis(
            window_size=window_length, groups=[[0], [1, 2], np.arange(3, window_length, 1)]
        )
        self.t1 = int(0.05 * sampling_rate_hz)
        assert self.t1 != 0  # based on approximation using an image in the original paper,
        # because not explicitly stated in the original paper

    def _detect_single_dataset(self, data) -> Dict[str, pd.DataFrame]:
        """Detect gait events for a single sensor data set and put into correct output stride list."""

        acc = data[BF_ACC]

        # find events
        event_detection_func = self._select_all_event_detection_method()
        ic, tc = event_detection_func(acc)
        event_list_ = self._get_event_list(ic, tc)
        event_list_ = pd.DataFrame(event_list_).set_index("s_id")
        return {"event_list_": event_list_}

    def _find_all_events(self, acc: pd.DataFrame) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """Find events in provided data by looping over single strides."""
        # from eargait.utils.helpers import butter_lowpass_filter
        # for col in acc.columns:
        #    acc[col] = butter_lowpass_filter(acc[col], 20, self.sampling_rate_hz/2, 4)
        acc_ssa = self._fit_transform_ssa(acc)
        ic, ic_sides = self._detect_ic(acc, acc_ssa)
        acc_ml_wo_trend = acc_ssa["acc_ml"][1] + acc_ssa["acc_ml"][2] + np.mean(acc["acc_ml"])
        tc, tc_sides = self._detect_tc(acc_ml_wo_trend, ic, ic_sides)

        ic += acc.index[0]
        tc += acc.index[0]

        return (
            pd.DataFrame.from_dict({"ic": ic, "side": ic_sides}),
            pd.DataFrame.from_dict({"tc": tc, "side": tc_sides}),
        )

    def _fit_transform_ssa(self, acc: pd.DataFrame) -> Dict:
        signal_compents = {}
        for axis in acc.columns:
            signal_compents[axis] = self.ssa.fit_transform(acc[axis].to_numpy().reshape(1, -1))
        return signal_compents

    def _detect_ic(self, acc: pd.DataFrame, acc_ssa: dict):
        ap_dominant = acc_ssa["acc_pa"][1]
        mins, _ = find_peaks(ap_dominant * -1, height=0, distance=self.sampling_rate_hz * 0.2)

        acc_wo_trend = {}
        for axes in acc_ssa.keys():
            acc_wo_trend[axes] = acc_ssa[axes][1] + acc_ssa[axes][2] + np.mean(acc[axes])

        # Note: Modification original paper.
        #       acc_si * -1 because defined as pointing upwards in paper, here: pointing downwards
        ic_events = np.array(
            [self._find_min_in_interval(acc_wo_trend["acc_pa"], acc_wo_trend["acc_si"] * -1, x, self.t1) for x in mins]
        )
        ic_side = []
        for i in range(len(ic_events) - 2):
            mean1 = np.mean(acc_wo_trend["acc_ml"][ic_events[i] : ic_events[i + 1]])
            mean2 = np.mean(acc_wo_trend["acc_ml"][ic_events[i + 1] : ic_events[i + 2]])

            # Note: Modification original paper.
            #       Flipped smaller sign because otherwise ipsi and contralateral are swapped
            if mean1 < mean2:
                ic_side.append("ipsilateral")
            else:
                ic_side.append("contralateral")
        ic_side = (
            ic_side + ["ipsilateral", "contralateral"]
            if ic_side[-1] == "contralateral"
            else ic_side + ["contralateral", "ipsilateral"]
        )
        return ic_events, np.array(ic_side)

    @staticmethod
    def _find_min_in_interval(acc_pa, acc_si, peak, t):
        if peak < t:
            t = peak
        elif (peak + t) > len(acc_pa):
            t = len(acc_pa) - peak
        interval = acc_pa[peak - t : peak + t] * (acc_si[peak - t : peak + t])
        return np.argmin(interval) + peak - t

    def _detect_tc(self, acc_ml, ic, ic_side):
        # contra TC: find local max BEFORE contra IC --> flip to local min
        # ipsi TC: find first local minimum AFFTER contra IC --> flip to local max
        # On ml axis without trend corrected with mean
        maximas, _ = find_peaks(acc_ml, width=self.sampling_rate_hz * 0.1)
        minimas, _ = find_peaks(acc_ml * -1, width=self.sampling_rate_hz * 0.1)
        tc_contra = []
        tc_ipsi = []
        for i in ic[ic_side == "contralateral"]:
            potenial_contra = maximas[(i - maximas) > 0]
            if potenial_contra.shape[0] > 0:
                tc_contra.append(potenial_contra[-1])
            else:
                tc_contra.append(np.nan)
            potential_ipsi = minimas[(minimas - i) > 0]
            if potential_ipsi.shape[0] > 0:
                tc_ipsi.append(potential_ipsi[0])
            else:
                tc_ipsi.append(np.nan)
        tc_contra = pd.DataFrame(data=tc_contra, columns=["tc"])
        tc_contra = tc_contra.assign(side="contralateral")
        tc_ipsi = pd.DataFrame(data=tc_ipsi, columns=["tc"])
        tc_ipsi = tc_ipsi.assign(side="ipsilateral")
        tc = pd.concat([tc_contra, tc_ipsi])
        tc = tc.sort_values(by="tc")
        return tc["tc"].to_numpy(), tc["side"].to_numpy()

    def plot(self: Self) -> Self:
        if not self._has_event_list():
            return None
        fig, axes = plt.subplots(2, 1, sharex=True)

        for i, sensor_pos in enumerate(self.event_list_.keys()):
            axes[i].plot(self.data[sensor_pos]["acc_si"], label="acc_si", color="b")
            axes[i].plot(self.data[sensor_pos]["acc_pa"], label="acc_pa", color="b", alpha=0.3, ls="--")
            axes[i].plot(self.data[sensor_pos]["acc_ml"], label="acc_ml", color="b", alpha=0.3, ls="-.")

            axes[i].set_title(sensor_pos)
            events = self.event_list_[sensor_pos]
            data = self.data[sensor_pos]

            for side, color in zip(["ipsilateral", "contralateral"], ["orange", "green"]):
                for event, marker in zip(["ic", "tc"], ["o", "v"]):
                    if event == "tc":
                        continue
                    tmp = events.loc[events.side == side, event].dropna()
                    axes[i].plot(tmp, data.loc[tmp]["acc_si"], marker, c=color, label=side + " " + event, ms=6, lw=1)
                    if event == "ic":
                        for ev in tmp:
                            axes[i].axvline(ev, c=color, lw=1, alpha=0.5)
        axes[0].legend(fontsize=12, bbox_to_anchor=(1.0, 0.0), loc="upper left")
        # plt.tight_layout()
        plt.show()
        return self



"""Base class for spatial parameter methods."""


class SpatialParamsBase:
    """Base Class for Spatial Parameter Estimation."""

    def __init__(self):
        pass

    def estimate(self, data, event_list):
        raise NotImplementedError("No valid spatial parameter method was provided.")



"""Mixed Diao and Jarchi (IC detection diao, side assignment jarchi). Not recommendet to use."""
from typing import Dict, Optional, Tuple, TypeVar, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from pyts.decomposition import SingularSpectrumAnalysis
from scipy.signal import argrelextrema, find_peaks, firwin, kaiserord, lfilter

from eargait.event_detection.base_event_detection import BaseEventDetection
from eargait.utils.consts import BF_ACC
from eargait.utils.helper_gaitmap import SensorData

Self = TypeVar("Self", bound="MixedEventDetection")

# pylint: skip-file


class MixedEventDetection(BaseEventDetection):
    """Find gait events in the IMU raw signal based on signal characteristics.

    Parameters
    ----------
    window_length


    Attributes
    ----------
    To Do

    Other Parameters
    ----------------
    To Do

    Examples
    --------
    Get gait events from single sensor signal

    >>> event_detection = MixedEventDetection()
    >>> event_detection.detect(data=data, sampling_rate_hz=200.0)
            ic      tc       side
    s_id
    0      651.0    584.0    ipsilateral
    1      839.0    802.0    contralateral
    2      1089.0   1023.0   ipsilateral
    ...

    """

    window_length: int
    segmented_event_list: Optional[Union[pd.DataFrame, Dict[str, pd.DataFrame]]]

    data: SensorData
    sampling_rate_hz: float
    ssa: SingularSpectrumAnalysis
    # stride_list: pd.DataFrame

    # filter parameter
    filter: bool
    filter_order_N: int
    filter_taps: np.ndarray
    filtered_data: SensorData

    def __init__(
        self,
        window_length: int = 200,
        sampling_rate_hz: Optional[float] = 200,
        filter: bool = False,
    ):
        super(MixedEventDetection, self).__init__(sampling_rate_hz)
        raise Warning("Mixed event detection method is not fully implemented and hence not recommended to use.")
        self.window_length = window_length
        self.filter = filter
        self.ssa = SingularSpectrumAnalysis(
            window_size=window_length, groups=[[0], [1], np.arange(2, window_length, 1)]
        )

        if filter:
            # initialize FIR filter
            nyq_rate = self.sampling_rate_hz / 2.0
            width = 2.0 / nyq_rate
            ripple_db = 60.0
            self.filter_order_N, beta = kaiserord(ripple_db, width)
            cutoff_hz = 5.0
            self.filter_taps = firwin(self.filter_order_N, cutoff_hz / nyq_rate, window=("kaiser", beta))

    def _detect_single_dataset(self, data) -> Dict[str, pd.DataFrame]:
        """Detect gait events for a single sensor data set and put into correct output stride list."""

        if self.filter:
            data = self._filter_data(data)

        acc = data[BF_ACC]

        # find events
        event_detection_func = self._select_all_event_detection_method()
        ic, tc = event_detection_func(acc, self.ssa, self.sampling_rate_hz)
        segmented_event_list = self._get_segmented_event_list(ic, tc)
        segmented_event_list = pd.DataFrame(segmented_event_list).set_index("s_id")
        return {"segmented_event_list": segmented_event_list}

    def _find_all_events(
        self, acc: pd.DataFrame, ssa: SingularSpectrumAnalysis, sampling_rate_hz: float
    ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """Find events in provided data by looping over single strides."""

        acc_si = acc["acc_si"].to_numpy().reshape(1, -1)
        acc_ssa_si = ssa.fit_transform(acc_si)
        acc_ml = acc["acc_ml"].to_numpy().reshape(1, -1) * -1  # changed
        acc_ssa_ml = ssa.fit_transform(acc_ml)

        ic, ic_sides = self._detect_ic(acc_ssa_si[1], acc_ssa_ml[2] + acc_ssa_ml[1], sampling_rate_hz)
        tc, tc_sides = self._detect_tc(acc_ssa_ml[2] + acc_ssa_ml[1], ic, ic_sides)

        ic += acc.index[0]
        tc += acc.index[0]

        return (
            pd.DataFrame.from_dict({"ic": ic, "side": ic_sides}),
            pd.DataFrame.from_dict({"tc": tc, "side": tc_sides}),
        )

    @staticmethod
    def _detect_ic(
        acc_si_dominant: np.ndarray, acc_ml_wo_trend: np.ndarray, sampling_rate_hz: float
    ) -> Tuple[np.ndarray, np.ndarray]:
        # find minimum on SI axis with removed trend. Peaks corresponds to IC
        peaks_ic, prop = find_peaks(acc_si_dominant * -1, height=0, distance=sampling_rate_hz * 0.2)
        # determine ipsilateral and contralateral IC
        ic_side = []
        for i in range(len(peaks_ic) - 2):
            mean1 = np.mean(acc_ml_wo_trend[peaks_ic[i] : peaks_ic[i + 1]])
            mean2 = np.mean(acc_ml_wo_trend[peaks_ic[i + 1] : peaks_ic[i + 2]])
            if mean1 > mean2:
                ic_side.append("ipsilateral")
            else:
                ic_side.append("contralateral")
        ic_side = (
            ic_side + ["ipsilateral", "contralateral"]
            if ic_side[-1] == "contralateral"
            else ic_side + ["contralateral", "ipsilateral"]
        )
        return peaks_ic, np.array(ic_side)

    @staticmethod
    def _detect_tc(acc_ml_wo_trend: np.ndarray, ic: np.ndarray, sides: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:

        mins = argrelextrema(acc_ml_wo_trend, np.less, order=2)[0]  # changes np.greater used to be np.less
        contra = ic[sides == "contralateral"]

        contra = contra[contra < mins[-1]]
        tc_ipsi = [int(mins[mins > x][0]) for x in contra]

        maxs = argrelextrema(acc_ml_wo_trend, np.greater, order=2)[0]  # changes  np.less used to be np.greater

        ipsi = ic[sides == "ipsilateral"]
        ipsi = ipsi[ipsi < maxs[-1]]
        tc_contra = [int(maxs[maxs > x][0]) for x in ipsi]

        tc_ipsi = list(set(tc_ipsi))
        tc_contra = list(set(tc_contra))

        # combine
        toe_off = np.array(tc_ipsi + tc_contra).astype(int)
        toe_off_side = np.array(["ipsilateral" for x in tc_ipsi] + ["contralateral" for x in tc_contra])

        # sort
        toe_off_side = toe_off_side[toe_off.argsort()]
        toe_off = toe_off[toe_off.argsort()]
        return toe_off, toe_off_side

    def _filter_data(self, data):
        delay = int(0.5 * (self.filter_order_N - 1))
        filtered_data = data.apply(self._apply_filter_on_column, axis=0).to_numpy()[delay:]
        index = data.index.to_numpy()[0 : data.shape[0] - delay]
        filtered_data = pd.DataFrame(filtered_data, index=index, columns=data.columns)
        return filtered_data

    def _apply_filter_on_column(self, x):
        return lfilter(self.filter_taps, 1.0, x)

    @staticmethod
    def plot_tmp(acc, acc_si_wo, acc_ml_wo):
        fig, axes = plt.subplots(1, 1)
        axes.plot(acc["acc_si"].to_numpy(), label="acc_si", c="red")
        axes.plot(acc["acc_ml"].to_numpy(), label="acc_ml", c="b")
        axes.plot(acc_si_wo[1], label="acc_si_wo_trend", c="red", ls="--")
        axes.plot(acc_ml_wo[1], label="acc_ml_wo_trend", c="b", ls="--")
        plt.legend()
        plt.show()

    def plot(self: Self) -> Self:
        fau_colors = {
            "fau": "#003865",  # dunkelblau
            "phil": "#c99313",  # gelb
            "wiso": "#8d1429",  # rot
            "med": "#00b1eb",  # hellblau
            "nat": "#009b77",  # grün
            "tech": "#98a4ae",  # grau
        }
        fau_blaues = [fau_colors["fau"], fau_colors["med"], fau_colors["tech"]]
        if not self._has_event_list():
            return None
        fig, axes = plt.subplots(2, 1, sharex=True, figsize=(12, 6))

        for i, sensor_pos in enumerate(self.segmented_event_list_.keys()):
            for axis, style, color in zip(["acc_si", "acc_ml", "acc_pa"], ["-", "--", "-."], fau_blaues):
                axes[i].plot(self.data[sensor_pos][axis], label=axis, color=color, ls=style)

            if self.filter:
                data = self._filter_data(self.data[sensor_pos])
                t = data["acc_si"].index.to_numpy()
                for axis, ls in zip(data.columns, ["--", "-", "-."]):
                    if axis == "acc_pa":
                        continue
                    ssa = self.ssa.fit_transform(data[axis].to_numpy().reshape(1, -1))
                    t = t[0 : len(ssa[0])]
                    if axis == "acc_si":
                        ssa = ssa[1]
                        label = axis + "_dom_os"
                    else:
                        ssa = ssa[1] + ssa[2]
                        label = axis + "_wo_trend"

                    axes[i].plot(t, ssa, label=label, color="red", alpha=1.0, ls=ls)

            axes[i].set_title(sensor_pos)
            strides = self.segmented_event_list_[sensor_pos]
            data = self.data[sensor_pos]

            for side, color in zip(["ipsilateral", "contralateral"], ["orange", "green"]):
                for event, marker in zip(["ic", "tc"], ["o", "v"]):
                    tmp = strides.loc[strides.side == side, event].dropna()
                    axes[i].plot(tmp, data.loc[tmp]["acc_si"], marker, c=color, label=side + " " + event, ms=6, lw=1)
                    if event == "ic":
                        for ev in tmp:
                            axes[i].axvline(ev, c=color, lw=1, alpha=0.5)

        axes[0].legend(fontsize=12, bbox_to_anchor=(1.0, 1.0), loc="upper left")
        plt.tight_layout()
        plt.subplots_adjust(hspace=0.138)
        plt.show()
        return self



"""The event detection algorithm by Diao et al, 2020 [1]."""
from typing import Dict, Tuple, TypeVar

import numpy as np
import pandas as pd
from pyts.decomposition import SingularSpectrumAnalysis
from scipy.signal import argrelextrema, find_peaks, firwin, kaiserord, lfilter

from eargait.event_detection.base_event_detection import BaseEventDetection
from eargait.utils.consts import BF_ACC
from eargait.utils.helper_gaitmap import SensorData

Self = TypeVar("Self", bound="DiaoEventDetection")


class DiaoEventDetection(BaseEventDetection):
    """Find gait events in the IMU raw signal based on signal characteristics.

    DiaoEventDetection uses a Singuar Spectrum Analysis to decompose acceleration data.
    The peaks in the dominant ozcillation of the SI axis are consindered to be the initial contact (IC).
    Ipsi and contralateral sides are assigned by looking at the adjecent sample in the ML axis.
    Terminate contacts (TC) are determined based the next minima or maxima in the ML axis.
    More information on the algorithm can be found in the paper by Diao et al. [1].

    Parameters
    ----------
    data
        The data passed to the `detect` method.
    sampling_rate_hz
        The sampling rate of the data
    ssa
        Singular Spectrum Analysis (SSA)
    window_length
        The window length for the SSA.


    Attributes
    ----------
    event_list_
        Event list or dictionary with such values.
        The result of the `detect` method holding all temporal gait events. A stride is defined
        such that terminate contact is follwed by an initial contact of the same foot.
        Gait sequences with no steady alternating sequence of ipsi- and contralateral strides are removed.
    nonconsistent_event_list_
        Event list with non consistent gait events. Non-consistent gait event are stride that do not follow the
        stride defintion as described in 'event_list_', e.g. initial contact is before the terminate contact.


    Examples
    --------
    Get gait events from single sensor signal

    >>> event_detection = DiaoEventDetection()
    >>> event_detection.detect(data=data, sample_rate_hz=200.0)
            ic      tc       side
    s_id
    0      651.0    584.0    ipsilateral
    1      839.0    802.0    contralateral
    2      1089.0   1023.0   ipsilateral
    ...

    Notes
    -----
    terminal contact (`tc`), often also referred to as toe off:
        At `tc` the movement of the ankle joint changes from a plantar flexion to a dorsal extension in the sagittal
        plane..

    initial contact (`ic`), often also referred to as heel strike:
        At `ic` the foot decelerates rapidly when the foot hits the ground.

    ipsilateral and contralateral:
        In contrast to foot worn sensors, ear worn sensors contain gait events from left and right foot.
        A stride of the foot situated at same body side as the ear worn sensor is referred to as ipsilateral stride.
        A stride of the foot situated at the oppisite body side if referred to as contralateral stride.
        A gait sequence is a steady alternating sequence of ipsi- and contralateral strides.


    [1] Diao, Y., Ma, Y., Xu, D., Chen, W., & Wang, Y. (2020). A novel gait parameter estimation method for healthy
    adults and postoperative patients with an ear-worn sensor. Physiological measurement, 41(5), 05NT01.

    """

    window_length: int
    ssa: SingularSpectrumAnalysis

    # filter parameter
    filter_order_n: int
    filter_taps: np.ndarray
    filtered_data: SensorData

    def __init__(
        self,
        sample_rate_hz: int,
        window_length: int = None,
    ):
        self.window_length = window_length
        self.ssa = None
        super().__init__(sample_rate_hz=sample_rate_hz)

    def _detect_single_dataset(self, data) -> Dict[str, pd.DataFrame]:
        """Detect gait events for a single sensor data set and put into correct output stride list."""
        self._initialize_filter()

        if self.window_length is None:
            self.window_length = self.sample_rate_hz
        assert self.window_length == self.sample_rate_hz

        if self.ssa is None:
            self._initialize_ssa()

        data = self._filter_data(data)

        acc = data[BF_ACC]
        assert acc.shape[0] > 3 * self.sample_rate_hz, "Walking Bout length must be greater than 3 seconds."

        # find events
        event_detection_func = self._select_all_event_detection_method()
        ic, tc = event_detection_func(acc, self.ssa, self.sample_rate_hz)
        event_list = self._get_event_list(ic, tc)
        event_list = pd.DataFrame(event_list).set_index("s_id")
        return {"event_list_": event_list}

    def _find_all_events(
        self, acc: pd.DataFrame, ssa: SingularSpectrumAnalysis, sample_rate_hz: float
    ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """Find events in provided data by looping over single strides."""
        acc_si = acc["acc_si"].to_numpy().reshape(1, -1)
        acc_ssa_si = ssa.fit_transform(acc_si)
        acc_ml = acc["acc_ml"].to_numpy().reshape(1, -1) * -1  # changed
        acc_ssa_ml = ssa.fit_transform(acc_ml)

        ic, ic_sides = self._detect_ic(acc_ssa_si[1], acc_ssa_ml[1] + acc_ssa_ml[2], sample_rate_hz)
        tc, tc_sides = self._detect_tc(acc_ssa_ml[1] + acc_ssa_ml[2], ic, ic_sides)

        ic += acc.index[0]
        tc += acc.index[0]

        return (
            pd.DataFrame.from_dict({"ic": ic, "side": ic_sides}),
            pd.DataFrame.from_dict({"tc": tc, "side": tc_sides}),
        )

    @staticmethod
    def _detect_ic(
        acc_si_dominant: np.ndarray, acc_ml_wo_trend: np.ndarray, sample_rate_hz: float
    ) -> Tuple[np.ndarray, np.ndarray]:
        # find minimum on SI axis with removed trend. Peaks corresponds to IC
        peaks_ic, _ = find_peaks(acc_si_dominant * -1, height=0.2, distance=sample_rate_hz * 0.2)
        # determine ipsilateral and contralateral IC
        sides = np.array(["contralateral" for i in range(peaks_ic.shape[0])])

        sides[(acc_ml_wo_trend[peaks_ic + 1] - acc_ml_wo_trend[peaks_ic]) > 0] = "ipsilateral"
        # flipped compared to Diao, because acc_ml was flipped in previous function
        return peaks_ic, sides

    @staticmethod
    def _detect_tc(acc_ml_wo_trend: np.ndarray, ic: np.ndarray, sides: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:

        mins = argrelextrema(acc_ml_wo_trend, np.less, order=2)[0]  # changes np.greater used to be np.less
        contra = ic[sides == "contralateral"]

        contra = contra[contra < mins[-1]]
        tc_ipsi = [int(mins[mins > x][0]) for x in contra]

        maxs = argrelextrema(acc_ml_wo_trend, np.greater, order=2)[0]  # changes  np.less used to be np.greater

        ipsi = ic[sides == "ipsilateral"]
        ipsi = ipsi[ipsi < maxs[-1]]
        tc_contra = [int(maxs[maxs > x][0]) for x in ipsi]

        tc_ipsi = list(set(tc_ipsi))
        tc_contra = list(set(tc_contra))

        # combine
        toe_off = np.array(tc_ipsi + tc_contra).astype(int)
        toe_off_side = np.array(["ipsilateral" for x in tc_ipsi] + ["contralateral" for x in tc_contra])

        # sort
        toe_off_side = toe_off_side[toe_off.argsort()]
        toe_off = toe_off[toe_off.argsort()]
        return toe_off, toe_off_side

    def _filter_data(self, data):
        delay = int(0.5 * (self.filter_order_n - 1))
        filtered_data = data.apply(self._apply_filter_on_column, axis=0).to_numpy()[delay:]
        index = data.index.to_numpy()[0 : data.shape[0] - delay]
        filtered_data = pd.DataFrame(filtered_data, index=index, columns=data.columns)
        return filtered_data

    def _apply_filter_on_column(self, x):
        return lfilter(self.filter_taps, 1.0, x)

    def _initialize_filter(self):
        # initialize FIR filter
        nyq_rate = self.sample_rate_hz / 2.0
        width = 2.0 / nyq_rate
        ripple_db = 60.0
        self.filter_order_n, beta = kaiserord(ripple_db, width)
        cutoff_hz = 5.0
        self.filter_taps = firwin(self.filter_order_n, cutoff_hz / nyq_rate, window=("kaiser", beta))

    def _initialize_ssa(self) -> Self:
        self.ssa = SingularSpectrumAnalysis(
            window_size=int(self.window_length), groups=[[0], [1], np.arange(2, self.window_length, 1)]
        )



"""Base class for all event detection algorithms based on ear worn motion sensors."""
import warnings
from typing import Callable, Dict, Optional, TypeVar, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from eargait.utils.helper_gaitmap import BaseEventDetection as BaseEventDetectionGaitmap
from eargait.utils.helper_gaitmap import (
    Hashable,
    SensorData,
    get_multi_sensor_names,
    invert_result_dictionary,
    is_sensor_data,
    set_params_from_dict,
)

Self = TypeVar("Self", bound="BaseEventDetection")


class BaseEventDetection(BaseEventDetectionGaitmap):
    """Base class for gait event detection methods.

    Parameters
    ----------
    data
        The IMU data passed to the `detect` method.
    sample_rate_hz
        The sample rate of the data


    Attributes
    ----------
    event_list_
        Event list or dictionary with such values.
        The result of the `detect` method holding all temporal gait events. A stride is defined
        such that terminate contact is follwed by an initial contact of the same foot.
    event_list_consistent_
        Event list with consistent gait events, hence, an alternating sequence of ipsi- and contralateral events.
        Non consistent (non-alternating) step are set to NaN.

    Examples
    --------
    Get gait events from single sensor signal

    >>> event_detection = BaseEventDetection()
    >>> event_detection.detect(data=data, sample_rate_hz=200.0)
            ic      tc       side
    s_id
    0      651.0    584.0    ipsilateral
    1      839.0    802.0    contralateral
    2      1089.0   1023.0   ipsilateral
    ...

    Notes
    -----
    terminal contact (`tc`), often also referred to as toe off:
        At `tc` the movement of the ankle joint changes from a plantar flexion to a dorsal extension in the sagittal
        plane..

    initial contact (`ic`), often also referred to as heel strike:
        At `ic` the foot decelerates rapidly when the foot hits the ground.

    ipsilateral and contralateral:
        In contrast to foot worn sensors, ear worn sensors contain gait events from left and right foot.
        An event of the foot situated at same body side as the ear worn sensor is referred to as ipsilateral event.
        An event of the foot situated at the oppisite body side if referred to as contralateral event.
        A gait sequence is a steady alternating sequence of ipsi- and contralateral events.

    """

    data: SensorData = None
    sample_rate_hz: float

    event_list_: Union[pd.DataFrame, Dict[str, pd.DataFrame]]
    event_list_consistent_: Optional[Union[pd.DataFrame, Dict[str, pd.DataFrame]]]

    bool_use_event_list_consistent: bool

    def __init__(
        self,
        sample_rate_hz: float,
        bool_use_event_list_consistent: bool = False,
    ):
        self.sample_rate_hz = sample_rate_hz
        self.bool_use_event_list_consistent = bool_use_event_list_consistent

    def detect(self: Self, data: SensorData) -> Self:  # noqa
        """Find gait events in data.

        Parameters
        ----------
        data
            The data set holding the imu raw data

        Returns
        -------
        self
            The class instance with all result attributes populated

        """
        dataset_type = is_sensor_data(data, frame="body")

        self.data = data

        if dataset_type == "single":
            results = self._detect_single_dataset(self.data)  # noqa
        else:
            results_dict: Dict[Hashable, Dict[str, pd.DataFrame]] = {}
            for sensor in get_multi_sensor_names(data):
                results_dict[sensor] = self._detect_single_dataset(self.data[sensor])  # noqa
            results = invert_result_dictionary(results_dict)

        set_params_from_dict(self, results, result_formatting=True)
        return self

    @property
    def event_list_consistent_(self: Self):
        """Consistent based on alternating ipsi- and contralateral steps, non-consistent steps are set to NaN."""
        assert self._has_event_list()
        dataset_type = is_sensor_data(self.data, frame="body")
        if dataset_type == "single":
            event_list_consistent_ = self._single_consistent_event_list(self.event_list_)
        else:
            event_list_consistent_: Dict[Hashable, Dict[str, pd.DataFrame]] = {}
            for sensor in get_multi_sensor_names(self.event_list_):
                event_list_consistent_[sensor] = self._single_consistent_event_list(self.event_list_[sensor])
        return event_list_consistent_

    def plot(self: Self, plot_ssa: bool = False):
        """Plot gait data and events."""
        assert self._has_event_list()
        if isinstance(self.event_list_, dict) == 1:
            for side in self.event_list_.keys():
                self._plot(sensor_pos=side, plot_ssa=plot_ssa)
        else:
            self._plot(sensor_pos=None, plot_ssa=plot_ssa)

    def _select_all_event_detection_method(self: Self) -> Callable:  # noqa: no-self-use
        """Select the function to calculate the all events.

        This is separate method to make it easy to overwrite by a subclass.
        """
        return self._find_all_events

    def _has_event_list(self: Self):
        if not hasattr(self, "event_list_"):
            warnings.warn("No event_list exists. Please run self.detect().")
            return False
        return True

    @staticmethod
    def _get_event_list(ic: pd.DataFrame, tc: pd.DataFrame) -> pd.DataFrame:
        """Create an event list based on a list of ic's and tc's.

        To each IC a suitable TC is assigned taking the sides into consideration.
        A stride is defined such that the TC is followed by an IC.

        Parameters
        ----------
        ic
            Dataframe containing all IC events and the corresponding side

        tc
            Dataframe containing all TC events and the corresponding side

        Returns
        -------
        event_list
            Dataframe containing gait events of IMU signal

        """
        event_list = ic.copy()
        event_list.ic = event_list.ic.astype(float)
        event_list.insert(1, "tc", np.nan)
        for side in event_list.side.unique():
            ic_side = ic.loc[ic.side == side].ic.to_numpy()
            tc_side = tc.loc[tc.side == side].tc.to_numpy()
            for i in range(ic_side.shape[0] - 1):
                tc_idx = tc_side[(ic_side[i] < tc_side) * (ic_side[i + 1] > tc_side)]
                if tc_idx.shape[0] == 1:
                    event_list.loc[event_list.ic == ic_side[i + 1], "tc"] = tc_idx[-1]

            # handle first ic
            first_tc = tc_side[tc_side < ic_side[0]]
            if len(first_tc) > 0:
                event_list.loc[event_list.ic == ic_side[0], "tc"] = first_tc[-1]
        event_list.insert(0, "s_id", np.arange(0, event_list.shape[0], 1))
        return event_list

    def _detect_single_dataset(self, data) -> Dict[str, pd.DataFrame]:  # noqa
        """Gait event detection on single data set.

        Parameters
        ----------
        data
            Data of single dataset

        Returns
        -------
        event_list_ as dict in the form of: {"event_list_": pd.DataFrame}
        """
        raise NotImplementedError("Needs to be implemented by child class.")

    def _plot(self: Self, sensor_pos: str = None, plot_ssa: bool = False) -> Self:  # noqa
        fau_blaues = ["#003865", "#00b1eb", "#98a4ae"]

        _, axes = plt.subplots(1, 1, sharex=True, figsize=(8, 4))

        if sensor_pos is None:
            data = self.data
        else:
            data = self.data[sensor_pos]
        filtered_data = self._filter_data(data)

        for axis, style, color in zip(["acc_si", "acc_ml", "acc_pa"], ["-", "--", "-."], fau_blaues):
            axes.plot(data[axis], label=axis, color=color, ls=style)
            axes.plot(filtered_data[axis], label=axis, color=color, ls=style, alpha=0.2)

        if plot_ssa:

            t = filtered_data["acc_si"].index.to_numpy()
            ssa_si = self.ssa.fit_transform(filtered_data["acc_si"].to_numpy().reshape(1, -1))
            ssa_ml = self.ssa.fit_transform(filtered_data["acc_ml"].to_numpy().reshape(1, -1))
            # t = t[0:len(ssa_si[0])]
            axes.plot(
                t,
                ssa_si[1] + filtered_data["acc_si"].mean(),
                label="acc_si_dominant_with_mean",
                color="red",
                alpha=0.7,
                ls="--",
            )
            axes.plot(
                t,
                ssa_ml[1] + filtered_data["acc_ml"].mean(),
                label="acc_ml_dominant_with_mean",
                color="red",
                alpha=0.7,
                ls="-.",
            )
            axes.plot(
                t,
                ssa_ml[1] + ssa_ml[2] + ssa_ml[3] + data["acc_ml"].mean(),
                label="acc_ml_rest",
                color="red",
                alpha=0.7,
                ls="--",
            )

        if sensor_pos is not None:
            axes.set_title(sensor_pos)
            events = self.event_list_[sensor_pos]
        else:
            events = self.event_list_

        for side, color in zip(["ipsilateral", "contralateral"], ["orange", "green"]):
            for event, marker in zip(["ic", "tc"], ["o", "v"]):
                tmp = events.loc[events.side == side, event].dropna()
                axes.plot(tmp, data.loc[tmp]["acc_si"], marker, c=color, label=side + " " + event, ms=6, lw=1)
        axes.legend(fontsize=12, loc="upper right")
        plt.show()
        return self

    @staticmethod
    def _single_consistent_event_list(event_list_):
        tmp = event_list_.copy()
        tmp["diff"] = 1
        tmp.loc[tmp.side == "contralateral", "diff"] = -1
        tmp["diff"] = tmp["diff"].diff()
        r = np.abs(np.diff(np.sign(tmp["diff"]))) == 2
        r[0] = True
        # counter = sum(~(r))
        r = np.append(r, True)
        event_list_.loc[~r, ["ic", "tc"]] = np.nan
        return event_list_



"""Eargait: The Gait and Movement Analysis Package unsing Ear-Worn IMU sensors."""
from eargait.spatial_params.spatial_params_example_class import SpatialParamsExample

__all__ = ["SpatialParamsExample"]



"""Example Class for Spatial Parameter Estimation methods."""
import warnings
from pathlib import Path
from typing import Dict, TypeVar, Union

import numpy as np
import pandas as pd

from eargait.spatial_params.spatial_params_base import SpatialParamsBase
from eargait.utils.helper_datatype import EventList
from eargait.utils.helper_gaitmap import SensorData

HERE = Path(__file__).parent

Self = TypeVar("Self", bound="SpatialParamsExample")


class SpatialParamsExample(SpatialParamsBase):
    """Spatial Parameter Estimation Class using a pretrained Random Forest."""

    sample_rate_hz: int

    def __init__(self, sample_rate_hz):
        self.sample_rate_hz = sample_rate_hz
        super().__init__()

    def estimate(self, data: SensorData, event_list: EventList) -> Union[Dict, pd.DataFrame]:
        """Caution: Example function! Sets all parameter to NaN."""
        warnings.warn(
            "Example class for spatial parameter estimation is used. No spatial parameters are calculated, "
            "step length and stride length are set to NaN."
        )
        spatial_params = event_list.copy()
        spatial_params = spatial_params.rename(columns={"ic": "step_length", "tc": "stride_length"})
        spatial_params.step_length = np.nan
        spatial_params.stride_length = np.nan
        return spatial_params
